# Initiate brdr
from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.utils import multipolygons_to_singles
from brdr.utils import write_geojson

aligner0 = Aligner()

# Load thematic data

aligner0.load_thematic_data_file(
    "../tests/testdata/multipolygon.geojson", "theme_identifier"
)
aligner0.dict_thematic = multipolygons_to_singles(aligner0.dict_thematic)
aligner0.load_thematic_data_dict(
    aligner0.dict_thematic,
)
# gebruik de actuele adp-percelen adp= administratieve percelen
aligner = Aligner()
aligner.load_thematic_data_dict(
    aligner0.dict_thematic,
)
aligner.load_reference_data_grb_actual(grb_type=GRBType.ADP, partition=1000)

# Example how to use the Aligner
# rel_dist = 2
# dict_results_by_distance = {}
# dict_results_by_distance[aligner.relevant_distance] = aligner.process_dict_thematic(
#     relevant_distance=rel_dist,
#     od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE,
# )
# aligner.export_results("output/")
# show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)
dict_series, dict_predicted, diffs = aligner.predictor()
fcs = aligner.get_predictions_as_geojson(series_dict=dict_predicted, formula=True)
aligner.export_results("output/")
write_geojson("output/predicted.geojson", fcs["result"])
write_geojson("output/predicted_diff.geojson", fcs["result_diff"])



import unittest
from datetime import date, timedelta

import numpy as np
from shapely import Polygon, from_wkt

from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import (
    get_last_version_date,
    is_grb_changed,
    get_geoms_affected_by_grb_change,
    evaluate,
    GRBActualLoader,
    GRBFiscalParcelLoader,
)
from brdr.loader import DictLoader
from brdr.utils import (
    get_series_geojson_dict,
)


class TestGrb(unittest.TestCase):
    def test_get_last_version_date(self):
        # Check if the result of the _buffer_neg_pos gives an equal geometry
        geom = Polygon([(0, 0), (0, 10), (10, 10), (10, 0)])
        out = get_last_version_date(geom)
        self.assertIsNone(out)
        geom = Polygon(
            [(170000, 170000), (170000, 170100), (170100, 170100), (170100, 170000)]
        )
        out = get_last_version_date(geom)
        self.assertIsInstance(out, date)

    def test_is_grb_changed(self):
        # Check if the result of the _buffer_neg_pos gives an equal geometry
        geom = Polygon([(0, 0), (0, 10), (10, 10), (10, 0)])
        out = is_grb_changed(geom)
        self.assertIsNone(out)
        geom = Polygon(
            [(170000, 170000), (170000, 170100), (170100, 170100), (170100, 170000)]
        )
        out = is_grb_changed(geom, grb_type=GRBType.ADP, date_start=date(2024, 7, 16))
        self.assertFalse(out)
        out = is_grb_changed(geom, grb_type=GRBType.ADP, date_start=date(2021, 7, 16))
        self.assertTrue(out)

    def test_is_grb_changed_outerborder(self):
        geom = Polygon(
            [(170000, 170000), (170000, 172000), (172000, 172000), (172000, 170000)]
        )
        out = is_grb_changed(
            geom,
            border_distance=0,
            grb_type=GRBType.ADP,
            date_start=date(2024, 7, 1),
        )
        self.assertTrue(out)
        out = is_grb_changed(
            geom,
            border_distance=10,
            grb_type=GRBType.ADP,
            date_start=date(2024, 7, 1),
        )
        self.assertFalse(out)

    def test_get_geoms_affected_by_grb_change_outerborder(self):
        thematic_dict = {
            "theme_id_1": Polygon(
                [(170000, 170000), (170000, 172000), (172000, 172000), (172000, 170000)]
            )
        }
        aligner = Aligner()
        aligner.load_thematic_data(DictLoader(thematic_dict))
        dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
            aligner=aligner,
            grb_type=GRBType.ADP,
            date_start=date.today() - timedelta(days=30),
            date_end=date.today(),
            one_by_one=False,
            border_distance=0,
        )
        assert len(dict_affected.keys()) > 0

        dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
            aligner=aligner,
            grb_type=GRBType.ADP,
            date_start=date.today() - timedelta(days=30),
            date_end=date.today(),
            one_by_one=False,
            border_distance=10,
        )
        assert len(dict_affected.keys()) == 0

    def test_get_geoms_affected_by_grb_change(self):
        thematic_dict = {
            "theme_id_1": from_wkt(
                "MultiPolygon (((174184.09476602054201066 171899.68933439542888664, "
                "174400.56834639035514556 171832.959863749332726, "
                "174388.65236948925303295 171770.99678386366576888, "
                "174182.10876987033407204 171836.13745758961886168, "
                "174184.88916448061354458 171873.07698598300339654, "
                "174184.09476602054201066 171899.68933439542888664)))"
            )
        }
        aligner = Aligner()
        aligner.load_thematic_data(DictLoader(thematic_dict))
        dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
            aligner=aligner,
            grb_type=GRBType.ADP,
            date_start=date.today() - timedelta(days=1),
            date_end=date.today(),
            one_by_one=True,
        )
        assert len(dict_affected.keys()) == 0

        dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
            aligner=aligner,
            grb_type=GRBType.ADP,
            date_start=date.today() - timedelta(days=1000),
            date_end=date.today(),
            one_by_one=True,
        )
        assert len(dict_affected.keys()) == 0
        thematic_dict2 = {
            "theme_id_2": from_wkt(
                "MultiPolygon (((174180.20077791667426936 171966.14649116666987538, "
                "174415.60530965600628406 171940.9636807945498731, "
                "174388.65236948925303295 171770.99678386366576888, "
                "174182.10876987033407204 171836.13745758961886168, "
                "174184.88916448061354458 171873.07698598300339654, "
                "174180.20077791667426936 171966.14649116666987538)))"
            )
        }
        aligner2 = Aligner()
        aligner2.load_thematic_data(DictLoader(thematic_dict2))
        dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
            aligner=aligner2,
            grb_type=GRBType.ADP,
            date_start=date.today() - timedelta(days=1000),
            date_end=date.today(),
            one_by_one=True,
        )
        assert len(dict_affected.keys()) > 0

    def test_get_geoms_affected_by_grb_change_bulk(self):
        thematic_dict = {
            "theme_id_1": from_wkt(
                "MultiPolygon (((174180.20077791667426936 171966.14649116666987538, "
                "174415.60530965600628406 171940.9636807945498731, "
                "174388.65236948925303295 171770.99678386366576888, "
                "174182.10876987033407204 171836.13745758961886168, "
                "174184.88916448061354458 171873.07698598300339654, "
                "174180.20077791667426936 171966.14649116666987538)))"
            )
        }
        aligner = Aligner()
        aligner.load_thematic_data(DictLoader(thematic_dict))
        dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
            aligner=aligner,
            grb_type=GRBType.ADP,
            date_start=date.today() - timedelta(days=1),
            date_end=date.today(),
            one_by_one=False,
        )
        assert len(dict_affected.keys()) == 0

        dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
            aligner=aligner,
            grb_type=GRBType.ADP,
            date_start=date.today() - timedelta(days=1000),
            date_end=date.today(),
            one_by_one=False,
        )
        assert len(dict_affected.keys()) > 0

    def test_evaluate(self):
        thematic_dict = {
            "theme_id_1": from_wkt(
                "MultiPolygon (((174180.20077791667426936 171966.14649116666987538, "
                "174415.60530965600628406 171940.9636807945498731, "
                "174388.65236948925303295 171770.99678386366576888, "
                "174182.10876987033407204 171836.13745758961886168, "
                "174184.88916448061354458 171873.07698598300339654, "
                "174180.20077791667426936 171966.14649116666987538)))"
            )
        }
        base_aligner = Aligner()
        base_aligner.load_thematic_data(DictLoader(thematic_dict))
        base_aligner.load_reference_data(
            GRBFiscalParcelLoader(aligner=base_aligner, year="2022", partition=1000)
        )
        base_process_result = base_aligner.process_dict_thematic(relevant_distance=1)
        thematic_dict_formula = {}
        thematic_dict_result = {}
        for key in base_process_result:
            thematic_dict_result[key] = base_process_result[key]["result"]
            thematic_dict_formula[key] = base_aligner.get_formula(
                thematic_dict_result[key]
            )
        aligner_result = Aligner()
        aligner_result.load_thematic_data(DictLoader(thematic_dict_result))
        dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
            aligner=aligner_result,
            grb_type=GRBType.ADP,
            date_start=date(2022, 1, 1),
            date_end=date.today(),
            one_by_one=False,
        )

        actual_aligner = Aligner()
        loader = DictLoader(dict_affected)
        actual_aligner.load_thematic_data(loader)
        loader = GRBActualLoader(
            grb_type=GRBType.ADP, partition=1000, aligner=actual_aligner
        )
        actual_aligner.load_reference_data(loader)
        series = np.arange(0, 200, 10, dtype=int) / 100
        dict_series, dict_predicted, diffs_dict = actual_aligner.predictor(series)

        dict_evaluated, prop_dictionary = evaluate(
            actual_aligner,
            dict_series,
            dict_predicted,
            thematic_dict_formula,
            threshold_area=5,
            threshold_percentage=1,
        )

        fc = get_series_geojson_dict(
            dict_evaluated,
            crs=actual_aligner.CRS,
            id_field=actual_aligner.name_thematic_id,
            series_prop_dict=prop_dictionary,
        )

        print(fc["result"])



import numpy as np

from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import GRBActualLoader
from brdr.loader import DictLoader
from brdr.utils import get_oe_dict_by_ids, dict_series_by_keys
from examples import show_map, plot_series

if __name__ == "__main__":
    # EXAMPLE to test the algorithm for erfgoedobject with relevant distance 0.2m and
    # od_strategy SNAP_ALL_SIDE

    # Initiate brdr
    aligner = Aligner()
    # Load thematic data & reference data
    # dict_theme = get_oe_dict_by_ids([206363], oetype='erfgoedobjecten')
    aanduidingsobjecten = range(1, 10)
    dict_theme = get_oe_dict_by_ids(
        aanduidingsobjecten, oetype="aanduidingsobjecten"
    )  # noqa
    loader = DictLoader(dict_theme)
    aligner.load_thematic_data(loader)
    loader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner)
    aligner.load_reference_data(loader)

    # RESULTS
    # rel_dist = 0.2
    # dict_results_by_distance = {}
    # #put resulting tuple in a dictionary
    # dict_results_by_distance[rel_dist] = aligner.process_dict_thematic(rel_dist,2)
    # aligner.export_results("output/")
    # show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)

    series = np.arange(0, 500, 20, dtype=int) / 100
    # predict which relevant distances are interesting to propose as resulting geometry
    dict_series, dict_predicted, diffs = aligner.predictor(
        relevant_distances=series, od_strategy=2, threshold_overlap_percentage=50
    )
    dict_predicted = dict_series_by_keys(dict_predicted)
    for key in dict_predicted.keys():
        diff = {key: diffs[key]}
        plot_series(series, diff)
        show_map(
            dict_predicted[key],
            {key: aligner.dict_thematic[key]},
            aligner.dict_reference,
        )



from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import get_collection_grb_actual, GRBActualLoader
from brdr.loader import GeoJsonFileLoader
from brdr.utils import polygonize_reference_data, geojson_to_dicts
from examples import show_map, print_formula

# example to test what happens if we combine borders
# (so thematic data can use both polygons)

# If we just combine sets of polygons (fe parcels and buildings) these polygons will
# overlap, and gives some unwanted/unexpected results. The reference data can be
# preprocessed with 'shapely-polygonize', to create a new set of non-overlapping
# polygons. These non-overlapping polygons can be used as reference-data to align the
# theme-data.


if __name__ == "__main__":
    # Initiate brdr
    aligner = Aligner()

    # Load thematic data & reference data
    loader = GeoJsonFileLoader(
        path_to_file="../tests/testdata/test_parcel_vs_building.geojson",
        id_property="theme_id",
    )
    aligner.load_thematic_data(loader)

    adploader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner)
    gbgloader = GRBActualLoader(grb_type=GRBType.GBG, partition=1000, aligner=aligner)
    collection_adp, name_reference_id_adp = get_collection_grb_actual(
        aligner.get_thematic_union(),
        grb_type=GRBType.ADP,
        partition=1000,
        date_start=None,
        date_end=None,
    )
    dict_adp, dict_adp_properties = geojson_to_dicts(
        collection_adp, name_reference_id_adp
    )

    collection_gbg, name_reference_id_gbg = get_collection_grb_actual(
        aligner.get_thematic_union(),
        grb_type=GRBType.GBG,
        partition=1000,
        date_start=None,
        date_end=None,
    )
    dict_gbg, dict_gbg_properties = geojson_to_dicts(
        collection_gbg, name_reference_id_gbg
    )

    dict_adp_gbg = dict_adp
    dict_adp_gbg.update(dict_gbg)  # combine 2 dictionaries
    # make a polygonized version of the reference data with non-overlapping polygons
    dict_ref = polygonize_reference_data(dict_adp_gbg)
    aligner.load_reference_data_dict(dict_ref)

    rel_dist = 2
    dict_results_by_distance = {}
    dict_results_by_distance[rel_dist] = aligner.process_dict_thematic(rel_dist, 4)
    aligner.export_results("output/")
    show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)
    print_formula(dict_results_by_distance, aligner)



import logging

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(message)s", datefmt="%d-%b-%y %H:%M:%S"
)


class Logger:
    def __init__(self, feedback=None):
        self.feedback = feedback

    def feedback_debug(self, text):
        if self.feedback is not None:
            # self.feedback.pushInfo(text)
            return
        logging.debug(text)

    def feedback_info(self, text):
        if self.feedback is not None:
            self.feedback.pushInfo(text)
            return
        logging.info(text)

    def feedback_warning(self, text):
        if self.feedback is not None:
            self.feedback.pushInfo(text)
            return
        logging.warning(text)



# Thresholds
# Area in m² for excluding candidate reference when overlap(m²) is smaller than the
# threshold
THRESHOLD_EXCLUSION_AREA = 0
# Percentage for excluding candidate reference when overlap(%) is smaller than the
# threshold
THRESHOLD_EXCLUSION_PERCENTAGE = 0

# Buffer parameters:
#   Explanation and examples:
#   https://shapely.readthedocs.io/en/stable/reference/shapely.buffer.html
#   https://postgis.net/docs/ST_Buffer.html
# Distance to limit a buffered corner (MITER-join-style parameter)
MITRE_LIMIT = 10
# Used in buffer-operations to define a quarter circle
QUAD_SEGMENTS = 5

# Correction-parameters (technical)

# Multiplication-factor used in OD-strategy 2 (SNAP-BOTH SIDED) when calculating
# OD-area to take into account
BUFFER_MULTIPLICATION_FACTOR = 1.01
# Threshold-value to exclude circles getting processed (perfect circle = 1) based on
# POLSPY-POPPER algorithm
THRESHOLD_CIRCLE_RATIO = 0.98
# Distance used in a pos_neg_buffer to remove slivers (technical correction)
CORR_DISTANCE = 0.01

# Download-settings: when extracting features by URL
# max buffer around thematic geometry to download reference parcels
MAX_REFERENCE_BUFFER = 10
# Limit used when extracting features by URL, using the feature API (f.e. from GRB)
DOWNLOAD_LIMIT = 10000

# default CRS:
DEFAULT_CRS = "EPSG:31370"  # BelgianLambert72

# MULTI_SINGLE_ID_SEPARATOR #separator to split multipolygon_ids to single polygons
MULTI_SINGLE_ID_SEPARATOR = "*$*"


# GRB_CONSTANTS
# URL of the OGC feature API of actual GRB to extract collections
GRB_FEATURE_URL = "https://geo.api.vlaanderen.be/GRB/ogc/features/collections"

# URL of the OGC feature API of GRB fiscal parcels (situation of 1st of January) to
# extract collections
GRB_FISCAL_PARCELS_URL = "https://geo.api.vlaanderen.be/Adpf/ogc/features/collections"

# Property-name of version_date
GRB_VERSION_DATE = "VERSDATUM"

# Property-name of id of GRB-parcels
GRB_PARCEL_ID = "CAPAKEY"

# Property-name of id of GRB-parcels
GRB_BUILDING_ID = "OIDN"

# Property-name of id of GRB-parcels
GRB_KNW_ID = "OIDN"



from brdr.aligner import Aligner
from brdr.enums import OpenbaarDomeinStrategy
from brdr.loader import GeoJsonFileLoader

if __name__ == "__main__":
    # Initiate brdr
    aligner = Aligner()
    # Load local thematic data and reference data
    loader = GeoJsonFileLoader(
        "../tests/testdata/themelayer_referenced.geojson", "id_theme"
    )
    aligner.load_thematic_data(loader)
    loader = GeoJsonFileLoader("../tests/testdata/reference_leuven.geojson", "capakey")
    aligner.load_reference_data(loader)
    # Example how to use the Aligner
    rel_dist = 1
    dict_results_by_distance = {
        aligner.relevant_distance: aligner.process_dict_thematic(
            relevant_distance=rel_dist,
            od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE,
        )
    }
    aligner.export_results("output/")
    # show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)



import logging
import os.path

import numpy as np
import requests
from geojson import Feature
from geojson import FeatureCollection
from geojson import dump
from shapely import GeometryCollection
from shapely import make_valid
from shapely import node
from shapely import polygonize
from shapely import unary_union
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry

from brdr.constants import MULTI_SINGLE_ID_SEPARATOR, DEFAULT_CRS, DOWNLOAD_LIMIT
from brdr.enums import DiffMetric
from brdr.geometry_utils import get_partitions, get_bbox
from brdr.typings import ProcessResult


def get_series_geojson_dict(
    series_dict: dict[float, dict[str, ProcessResult]],
    crs: str,
    id_field: str,
    series_prop_dict: dict[float, dict[str, any]] = None,
    geom_attributes=True,
):
    """
    Convert a series of process results to a GeoJSON feature collections.
    """
    features_list_dict = {}

    for relative_distance, results_dict in series_dict.items():
        prop_dict = dict(series_prop_dict or {}).get(relative_distance, {})
        for theme_id, process_result in results_dict.items():
            properties = prop_dict.get(theme_id, {})
            properties[id_field] = theme_id
            properties["relevant_distance"] = relative_distance

            for results_type, geom in process_result.items():
                if results_type not in features_list_dict:
                    features_list_dict[results_type] = []

                feature = feature_from_geom(geom, properties, geom_attributes)
                features_list_dict[results_type].append(feature)

    crs_geojson = {"type": "name", "properties": {"name": crs}}
    return {
        result_type: FeatureCollection(features, crs=crs_geojson)
        for result_type, features in features_list_dict.items()
    }


def feature_from_geom(
    geom: BaseGeometry,
    properties: dict = None,
    geom_attributes=True,
) -> Feature:
    """
    Convert a geometry to a GeoJSON feature.

    Args:
        geom (BaseGeometry): The geometry to convert.
        properties (dict): The properties to include in the feature.
        geom_attributes (bool): Whether to include geometry attributes (default True).

    Returns:
        Feature: The GeoJSON feature.
    """

    properties = dict(properties or {})
    if geom_attributes:
        area = geom.area
        perimeter = geom.length
        properties["area"] = area
        properties["perimeter"] = perimeter
        properties["shape_index"] = perimeter / area if area != 0 else -1
    return Feature(geometry=geom, properties=properties)


def geojson_from_dict(dictionary, crs, id_field, prop_dict=None, geom_attributes=True):
    """
    get a geojson (featurecollection) from a dictionary of ids(keys) and geometries
    (values)
    """
    features = []
    for key, geom in dictionary.items():
        properties = dict(prop_dict or {}).get(key, {})
        properties[id_field] = key
        features.append(feature_from_geom(geom, properties, geom_attributes))
    crs_geojson = {"type": "name", "properties": {"name": crs}}
    geojson = FeatureCollection(features, crs=crs_geojson)
    return geojson


def write_geojson(path_to_file, geojson):
    parent = os.path.dirname(path_to_file)
    os.makedirs(parent, exist_ok=True)
    with open(path_to_file, "w") as f:
        dump(geojson, f)


def multipolygons_to_singles(dict_geoms):
    """
    Converts a dictionary of shapely-geometries to a dictionary containing only single
    polygons.

    This function iterates through a dictionary where values are Shapely-geometries
    and performs the following:

    * **Polygons:** Preserves the key and geometry from the original dictionary.
    * **MultiPolygons with one polygon:** Preserves the key and extracts the single
        polygon.
    * **MultiPolygons with multiple polygons:**
        * Creates new keys for each polygon by appending a suffix (_index) to the
            original key.
        * Assigns the individual polygons from the MultiPolygon to the newly created
            keys.

    Args:
        dict_geoms (dict): A dictionary where keys are identifiers and values are
            GeoJSON geometries.

    Returns:
        dict: A new dictionary containing only single polygons (as Polygon geometries).
              Keys are created based on the logic described above.

    Notes:
        * Geometries that are not Polygons or MultiPolygons are excluded with a warning
            message printed.
    """
    resulting_dict_geoms = {}
    for key in dict_geoms:
        geom = dict_geoms[key]
        if str(geom.geom_type) == "Polygon":
            resulting_dict_geoms[key] = geom
        elif str(geom.geom_type) == "MultiPolygon":
            polygons = list(geom.geoms)
            if len(polygons) == 1:
                resulting_dict_geoms[key] = polygons[0]
                continue
            i = 0
            for p in polygons:
                new_key = str(key) + MULTI_SINGLE_ID_SEPARATOR + str(i)
                resulting_dict_geoms[new_key] = p
                i = i + 1
        else:
            logging.debug("geom excluded: " + str(geom) + " for key: " + str(key))
    return resulting_dict_geoms


def polygonize_reference_data(dict_ref):
    """
    Creates a new dictionary with non-overlapping polygons based on a reference data
    dictionary.

    This function is designed to handle situations where the original reference data
    dictionary might contain:

    * Overlapping polygons: It creates new, non-overlapping polygons by combining all
        reference borders.
    * Multiple overlapping references: This function is useful when combining references
        like parcels and buildings that might overlap.

    **Important:** The original reference IDs are lost in the process of creating new
        non-overlapping polygons. New unique keys are assigned instead.

    Args:
        dict_ref (dict): A dictionary where keys are identifiers and values are Shapely
            geometries (assumed to be Polygons or MultiPolygons).

    Returns:
        dict: A new dictionary containing non-overlapping polygons derived from the
              original reference data.
              Keys are unique strings (reference IDs are lost).

    Notes:
        * Geometries that are not Polygons or MultiPolygons are excluded with a warning
            message printed.
    """
    arr_ref = []
    for key in dict_ref:
        arr_ref.append((make_valid(dict_ref[key])))
    arr_ref = node(GeometryCollection(arr_ref))
    polygonized = polygonize(arr_ref.geoms)
    dict_ref = {}
    i = 0
    for g in polygonized.geoms:
        if str(g.geom_type) in ["Polygon", "MultiPolygon"]:
            i = i + 1
            key = str(i)  # unique keys will be added (reference_id is lost)
            dict_ref[key] = make_valid(g)
        else:
            logging.warning("geom excluded: " + str(g))
    return dict_ref


def get_oe_dict_by_ids(objectids, oetype="aanduidingsobjecten"):
    """
    Fetches thematic data for a list of objectIDs from the Inventaris Onroerend Erfgoed
    API.

    This function retrieves information about designated heritage objects
    (erfgoedobjecten or aanduidingsobjecten) from the Flemish Agency for Heritage (
    Inventaris Onroerend Erfgoed) based on a list of their IDs.

    Args:
        objectids (list): A list of objectIDs of 'erfgoedobjecten' or
            'aanduidingsobjecten'.
        oetype (string): A string: 'aanduidingsobjecten' (default) or 'erfgoedobjecten'

    Returns:
        dict: A dictionary where keys are objectIDs (as strings) and values are
              GeoJSON geometry objects. If an erfgoedobject/aanduidingsobject is not
              found, a corresponding warning message will be logged, but it won't be\
              included in the returned dictionary.

    Raises:
        requests.exceptions.RequestException: If there is an error fetching data from
            the API.
    """
    dict_thematic = {}
    base_url = "https://inventaris.onroerenderfgoed.be/" + oetype + "/"
    headers = {"Accept": "application/json"}
    for a in objectids:
        url = base_url + str(a)
        response = requests.get(url, headers=headers).json()
        if "id" in response.keys():
            key = str(response["id"])
            geom = shape(response["locatie"]["contour"])
            dict_thematic[key] = geom
        else:
            logging.warning("object id " + str(a) + " not available in " + oetype)
    return dict_thematic


def get_oe_geojson_by_bbox(bbox, limit=1000):
    """
    Fetches GeoJSON data for designated heritage objects (aanduidingsobjecten) within
    a bounding box.

    This function retrieves information about aanduidingsobjecten from the Flemish
    Mercator public WFS service using a bounding box (bbox) as a filter. The bbox should
    be provided in the format "xmin,ymin,xmax,ymax" (EPSG:31370 projection).

    Args:
        bbox (str): A comma-separated string representing the bounding box in EPSG:31370
                   projection (e.g., "100000,500000,200000,600000").
        limit (int, optional): The maximum number of features to retrieve per request.
                              Defaults to 1000.

    Returns:
        dict: A dictionary containing the retrieved GeoJSON feature collection. This
              collection might be truncated if the total number of features exceeds
              the specified limit.
    """
    theme_url = (
        "https://www.mercator.vlaanderen.be/raadpleegdienstenmercatorpubliek/wfs?"
        "SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&"
        "TYPENAMES=ps:ps_aandobj&"
        f"COUNT={str(limit)}&"
        "SRSNAME=urn:ogc:def:crs:EPSG::31370&"
        f"BBOX={bbox}&outputFormat=application/json"
    )
    start_index = 0
    collection = {}
    while True:
        url = theme_url + "&startIndex=" + str(start_index)
        feature_collection = requests.get(url).json()
        if (
            "features" not in feature_collection
            or len(feature_collection["features"]) == 0
        ):
            break
        start_index = start_index + limit
        collection = collection | feature_collection
    return collection


def get_breakpoints_zerostreak(x, y):
    """
    Determine the extremes and zero_streaks of a graph based on the derivative, and
    return:
    * the breakpoints: extremes (breakpoints) of graph where 'change' occurs
    * the zero_streaks: ranges where the derivative is zero, ranges of relevant_distance
        where 'no-change' occurs

    Parameters:
    x (numpy.ndarray): The x values of the graph.
    derivative (numpy.ndarray): The y values of the graph.

    Returns:
    extremes: A list of tuples for breakpoints:
        *   relevant distance where extreme occurs
        *   extreme value
        *   minimum or maximum
    zero_streaks: A list of tuples for zero_streaks:
        * relevant distance where zero_streak starts
        * relevant distance where zero_streak ends
        * center of start- and end- zero_streak
        * counter of #relevant_distances where zero-streak holds on
        * extreme value for zero_streak
    """
    derivative = numerical_derivative(x, y)
    # plt.plot(x, y, label="y")
    # plt.plot( x, derivative, label="derivative")
    # plt.legend()
    # plt.show()
    extremes = []
    zero_streaks = []
    start_streak = None
    streak = 0
    write_zero_streak = False
    last_extreme = 0
    for i in range(1, len(x)):
        if round(derivative[i], 2) == 0:  # noqa
            streak = streak + 1
            if start_streak is None:
                start_streak = x[i - 1]
        elif streak != 0:
            write_zero_streak = True
        if start_streak is not None and (write_zero_streak or len(x) - 1 == i):
            end_streak = x[i - 1]
            if derivative[i] == 0:
                end_streak = x[i]
            center_streak = start_streak + (end_streak - start_streak) / 2
            zero_streaks.append(
                (start_streak, end_streak, center_streak, streak, last_extreme)
            )
            streak = 0
            start_streak = None
            write_zero_streak = False
            logging.debug("end_streak")
        if (
            i < len(x) - 1
            and round(derivative[i], 2) > 0  # noqa
            and derivative[i - 1] <= derivative[i]
            and derivative[i] >= derivative[i + 1]
        ):
            last_extreme = derivative[i]
            extremes.append((x[i], derivative[i], "maximum"))
        if (
            i < len(x) - 1
            and round(derivative[i], 2) < 0  # noqa
            and derivative[i - 1] >= derivative[i]
            and derivative[i] <= derivative[i + 1]
        ):
            last_extreme = derivative[i]
            extremes.append((x[i], derivative[i], "minimum"))
    for extremum in extremes:
        logging.debug(
            f"breakpoints: relevant_distance:"
            f"{extremum[0]:.2f}, extreme:{extremum[1]:.2f} ({extremum[2]})"
        )
    for st in zero_streaks:
        logging.debug(
            f"zero_streaks: [{st[0]:.2f} - {st[1]:.2f}] - center:{st[2]:.2f}"
            f" - counter:{st[3]:.2f} - min/max-extreme:{st[4]:.2f} "
        )
    # plt.plot(series, afgeleide, label='afgeleide-' + str(key))
    return extremes, zero_streaks


def numerical_derivative(x, y):
    """
    Calculate the numerical derivative of a graph.

    Parameters:
      x (numpy.ndarray): The x values of the graph.
      y (numpy.ndarray): The y values of the graph.

    Returns:
      numpy.ndarray: The derivative of y with respect to x.
    """

    dx = np.diff(x)
    dy = np.diff(y)
    derivative = dy / dx
    derivative = np.insert(derivative, 0, 0)
    # derivative = np.append(derivative, 0)

    return derivative


def filter_dict_by_key(dictionary, filter_key):
    """
    Filters a dictionary to only include keys matching a specific value.

    This function creates a new dictionary containing entries from the original
    dictionary where the key matches the provided `filter_key`.

    Args:
        dictionary (dict): The dictionary to filter.
        filter_key (str): The key value to filter by.

    Returns:
        dict: A new dictionary containing only entries where the key matches the
            `filter_key`.
    """
    return {key: dictionary[key] for key in dictionary.keys() if key == filter_key}


def diffs_from_dict_series(
    dict_series: dict[float, dict[str, ProcessResult]],
    dict_thematic: dict[str, BaseGeometry],
    diff_metric: DiffMetric = DiffMetric.CHANGES_AREA,
):
    """
    Calculates a dictionary containing difference metrics for thematic elements based on
     a distance series.

    This function analyzes the changes in thematic elements (represented by
    thematic_ids in `dict_thematic`) across different distances provided in the
    `dict_series`. It calculates a difference metric for each thematic element at
    each distance and returns a dictionary summarizing these differences.

    Args:
        dict_series (dict): A dictionary where thematic_ids are distances and
        values are tuples of two dictionaries.
                             - The first dictionary in the tuple represents thematic
                                element areas for a specific distance.
                             - The second dictionary represents the difference in
                                areas from the original thematic data for a specific
                                distance.
        dict_thematic (dict): A dictionary where thematic_ids are thematic element
            identifiers and values are GeoJSON geometry objects representing the
            original thematic data.
       diff_metric (DiffMetric): The metric used to determine the difference between
            the thematic and reference data.

    Returns:
        dict: A dictionary containing difference metrics for each thematic element
        (`key`) across different distances.
             The structure is as follows:
             {
                'thematic_key1': {
                    distance1: difference_metric1,
                    distance2: difference_metric2,
                    ...
                },
                'thematic_key2': {
                    distance1: difference_metric1,
                    distance2: difference_metric2,
                    ...
                },
                ...
             }

             - `difference_metric`: This value depends on the chosen calculation for
                thematic element change. The docstring provides examples like area
                difference, percentage change, and absolute difference.

    Raises:
        KeyError: If a thematic element key is missing from the results in
        `dict_series`.
    """
    thematic_ids = dict_thematic.keys()

    diffs = {thematic_id: {} for thematic_id in thematic_ids}

    # all the relevant distances used to calculate the series
    for rel_dist, results_dict in dict_series.items():

        for thematic_id in thematic_ids:
            result = results_dict.get(thematic_id, {}).get("result")
            result_diff = results_dict.get(thematic_id, {}).get("result_diff")
            # result_diff_plus = results_dict.get(thematic_id, {})\
            # .get("result_diff_plus")
            # result_diff_min = results_dict.get(thematic_id, {})\
            # .get("result_diff_min")

            diff = 0
            if (
                result_diff is None
                or result_diff.is_empty
                or result is None
                or result.is_empty
            ):
                diff = 0
            elif diff_metric == DiffMetric.TOTAL_AREA:
                diff = result.area - dict_thematic[thematic_id].area
            elif diff_metric == DiffMetric.TOTAL_PERCENTAGE:
                diff = result.area - dict_thematic[thematic_id].area
                diff = diff * 100 / result.area
            elif diff_metric == DiffMetric.CHANGES_AREA:
                # equals the symmetrical difference, so equal to
                # result_diff_plus.area + result_diff_min.area
                diff = result_diff.area
                # diff = result_diff_plus.area + result_diff_min.area
            elif diff_metric == DiffMetric.CHANGES_PERCENTAGE:
                diff = result_diff.area
                diff = diff * 100 / result.area

            # round, so the detected changes are within 10cm² or 0.1%
            diff = round(diff, 1)
            diffs[thematic_id][rel_dist] = diff

    return diffs


def get_collection(ref_url, limit):
    """
    Fetches a collection of features from a paginated API endpoint.

    This function retrieves a collection of features from a URL that supports
    pagination using a `startIndex` parameter. It iteratively retrieves features in
    chunks of the specified `limit` until no more features are available.

    Args:
        ref_url (str): The base URL of the API endpoint.
        limit (int): The maximum number of features to retrieve per request.

    Returns:
        dict: A dictionary representing the complete GeoJSON feature collection.
            This might be truncated if the total number of features exceeds the
            limitations of the API or server.

    Logs:
        - Debug logs the URL being used for each request during pagination.
    """
    start_index = 0
    collection = {}
    while True:
        url = ref_url + "&startIndex=" + str(start_index)
        logging.debug("called url: " + url)
        json = requests.get(url).json()
        feature_collection = dict(json)
        if (
            "features" not in feature_collection
            or len(feature_collection["features"]) == 0
        ):
            break
        start_index = start_index + limit
        if collection == {}:
            collection = feature_collection
        else:
            collection["features"].extend(feature_collection["features"])  # noqa
        if len(feature_collection["features"]) < limit:
            break
    return collection


def geojson_to_dicts(collection, id_property):
    data_dict = {}
    data_dict_properties = {}
    if collection is None or "features" not in collection:
        return data_dict, data_dict_properties
    for f in collection["features"]:
        key = str(f["properties"][id_property])
        geom = shape(f["geometry"])
        data_dict[key] = make_valid(geom)
        data_dict_properties[key] = f["properties"]
    return data_dict, data_dict_properties


def get_collection_by_partition(
    url, geometry, partition=1000, limit=DOWNLOAD_LIMIT, crs=DEFAULT_CRS
):
    collection = {}
    if geometry is None:
        collection = get_collection(
            _add_bbox_to_url(url=url, crs=crs, bbox=None), limit
        )
    elif partition < 1:
        collection = get_collection(
            _add_bbox_to_url(url=url, crs=crs, bbox=get_bbox(geometry)), limit
        )
    else:
        geoms = get_partitions(geometry, partition)
        for g in geoms:
            coll = get_collection(
                _add_bbox_to_url(url=url, crs=crs, bbox=get_bbox(g)), limit
            )
            if collection == {}:
                collection = dict(coll)
            elif "features" in collection and "features" in coll:
                collection["features"].extend(coll["features"])
    return collection


def _add_bbox_to_url(url, crs=DEFAULT_CRS, bbox=None):
    # Load the Base reference data
    if bbox is not None:
        url = url + "&bbox-crs=" + crs + "&bbox=" + bbox
    return url


def merge_dict_series(
    dict_series: dict[float, dict[str, ProcessResult]]
) -> dict[float, dict[str, ProcessResult]]:
    """
    Merges dict_series (dict_predicted) with  seperated IDs (MULTI_SINGLE_ID_SEPARATOR)
     to their original unique ID
    """
    dict_series_merged = {}
    for dist, item in dict_series.items():
        dict_series_merged[dist] = merge_process_results(item)
    return dict_series_merged


def merge_dict(dictionary: dict[str, BaseGeometry]) -> dict[str, BaseGeometry]:
    """
    Merges dict_series (dict_predicted) with  seperated IDs (MULTI_SINGLE_ID_SEPARATOR)
     to their original unique ID
    """
    out_dictionary = {}
    for id_theme, item in dictionary.items():
        id_theme_global = id_theme.split(MULTI_SINGLE_ID_SEPARATOR)[0]
        if id_theme_global not in out_dictionary:
            out_dictionary[id_theme_global] = [item]
        else:
            out_dictionary[id_theme_global].append(item)
    return {k: make_valid(unary_union(v)) for k, v in out_dictionary.items()}


def merge_process_results(
    result_dict: dict[str, ProcessResult]
) -> dict[str, ProcessResult]:
    """
    Merges geometries in a dictionary from multiple themes into a single theme.

    Args: result_dict (dict): A dictionary where keys are theme IDs and values are
        process results

    Returns: dict: A new dictionary with merged geometries, where keys are global
        theme IDs and values are merged geometries.

    """
    grouped_results: dict[str, ProcessResult] = {}

    for id_theme, process_result in result_dict.items():
        id_theme_global = id_theme.split(MULTI_SINGLE_ID_SEPARATOR)[0]
        if id_theme_global not in grouped_results:
            grouped_results[id_theme_global] = process_result
        else:
            for key in process_result:
                geom: BaseGeometry = process_result[key]  # noqa
                if geom.is_empty or geom is None:
                    continue
                existing: BaseGeometry = grouped_results[id_theme_global][key]  # noqa
                grouped_results[id_theme_global][key] = unary_union(  # noqa
                    [existing, geom]
                )

    return grouped_results


def processresult_to_dicts(dict_processresult):
    """
    Transforms a dictionary with all ProcessResults to individual dictionaries of the
    results
    Args:
        dict_processresult:

    Returns:

    """
    results = {}
    results_diff = {}
    results_diff_plus = {}
    results_diff_min = {}
    results_relevant_intersection = {}
    results_relevant_diff = {}
    for key in dict_processresult:
        processresult = dict_processresult[key]
        results[key] = processresult["result"]
        results_diff[key] = processresult["result_diff"]
        results_diff_plus[key] = processresult["result_diff_plus"]
        results_diff_min[key] = processresult["result_diff_min"]
        results_relevant_intersection[key] = processresult[
            "result_relevant_intersection"
        ]
        results_relevant_diff[key] = processresult["result_relevant_diff"]

    return (
        results,
        results_diff,
        results_diff_plus,
        results_diff_min,
        results_relevant_intersection,
        results_relevant_diff,
    )


def dict_series_by_keys(dict_series):
    """
    Transforms a dict_series into a dictionary with theme_id as keys, and a dictionary
    with all predicted distances and their resulting geometry as a value.
    Args:
        dict_series: a dictionary result of the 'series/predictor'

    Returns: dictionary with theme_id as keys, and a dictionary with all serial
    distances and their resulting geometry as a value.

    """
    dict_series_keys = {}
    for dist, res in dict_series.items():
        for key in res.keys():
            if key not in dict_series_keys.keys():
                dict_series_keys[key] = {}
            dict_series_keys[key][dist] = {key: res[key]}
    return dict_series_keys



import numpy as np

from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.utils import get_oe_dict_by_ids, write_geojson, dict_series_by_keys
from examples import show_map, plot_series

if __name__ == "__main__":
    # EXAMPLE to test the algorithm for erfgoedobject with relevant distance 0.2m and
    # od_strategy SNAP_ALL_SIDE

    # Initiate brdr
    aligner = Aligner()
    # Load thematic data & reference data
    # dict_theme = get_oe_dict_by_ids([206363], oetype='erfgoedobjecten')

    erfgoedobjecten = [
        # 206407,
        # 206403,
        # 206372,
        # 206369,
        # 206377,
        # 206371,
        # 206370,
        # 206368,
        206786
    ]
    dict_theme = get_oe_dict_by_ids(erfgoedobjecten, oetype="erfgoedobjecten")
    aligner.load_thematic_data_dict(dict_theme)
    aligner.load_reference_data_grb_actual(grb_type=GRBType.ADP, partition=1000)

    # RESULTS
    # rel_dist = 0.2
    # dict_results_by_distance = {}
    # #put resulting tuple in a dictionary
    # dict_results_by_distance[rel_dist] = aligner.process_dict_thematic(rel_dist,2)
    # aligner.export_results("output/")
    # show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)

    series = np.arange(0, 200, 20, dtype=int) / 100
    # predict which relevant distances are interesting to propose as resulting geometry
    dict_series, dict_predicted, diffs = aligner.predictor(
        relevant_distances=series, od_strategy=2, threshold_overlap_percentage=50
    )
    fcs = aligner.get_predictions_as_geojson(series_dict=dict_predicted)
    write_geojson("output/predicted.geojson", fcs["result"])
    write_geojson("output/predicted_diff.geojson", fcs["result_diff"])

    dict_predicted = dict_series_by_keys(dict_predicted)
    for key in dict_predicted.keys():
        diff = {key: diffs[key]}
        plot_series(series, diff)
        show_map(
            dict_predicted[key],
            {key: aligner.dict_thematic[key]},
            aligner.dict_reference,
        )



from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import GRBActualLoader
from brdr.loader import DictLoader
from brdr.utils import get_oe_dict_by_ids
from examples import print_formula
from examples import show_map

if __name__ == "__main__":
    # EXAMPLE for a thematic Polygon (aanduid_id 131635)

    # Initiate brdr
    aligner = Aligner()
    # Load thematic data & reference data
    dict_theme = get_oe_dict_by_ids([131635])
    loader = DictLoader(dict_theme)
    aligner.load_thematic_data(loader)
    loader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner)
    aligner.load_reference_data(loader)

    # RESULTS
    rel_dist = 2
    dict_results_by_distance = {rel_dist: aligner.process_dict_thematic(rel_dist, 4)}
    # put resulting tuple in a dictionary
    aligner.export_results("output/", formula=True)

    show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)
    print_formula(dict_results_by_distance, aligner)



import json
from abc import ABC

import requests as requests
from shapely.geometry.base import BaseGeometry

from brdr.typings import FeatureCollection
from brdr.utils import geojson_to_dicts


class Loader(ABC):
    def __init__(self):
        self.data_dict: dict[str, BaseGeometry] = {}
        self.data_dict_properties: dict[str, dict] = {}
        self.data_dict_source: dict[str, str] = {}

    def load_data(self):
        return self.data_dict, self.data_dict_properties, self.data_dict_source


class DictLoader(Loader):
    def __init__(self, data_dict: dict[str:BaseGeometry]):
        super().__init__()
        self.data_dict = data_dict

    def load_data(self):
        # self._prepare_reference_data()
        return super().load_data()


class GeoJsonLoader(Loader):
    def __init__(
        self,
        *,
        id_property: str = None,
        _input: FeatureCollection = None,
        # data_dict_properties=None, TODO?
    ):
        super().__init__()
        self.id_property = id_property
        self.input = _input

    def load_data(self):
        self._load_geojson_data()
        return super().load_data()

    def _load_geojson_data(self):
        """
        Load geometries of a GeoJSON and stores them in a dictionary.

        This method processes the thematic data from the input GeoJSON file. It
        iterates through each feature, extracts the relevant properties, converts the
        geometry to a valid shape, and stores it in a dictionary.

        Returns:
            None.
        """
        # THEMATIC PREPARATION
        self.data_dict, self.data_dict_properties = geojson_to_dicts(
            self.input, self.id_property
        )
        return


class GeoJsonFileLoader(GeoJsonLoader):
    def __init__(self, path_to_file, id_property):
        with open(path_to_file, "r") as f:
            _input = json.load(f)
        super().__init__(_input=_input, id_property=id_property)


class GeoJsonUrlLoader(GeoJsonLoader):
    def __init__(self, url, id_property):
        _input = requests.get(url).json()
        super().__init__(_input=_input, id_property=id_property)



import json
import logging
import os
from collections import defaultdict
from datetime import datetime
from math import pi
from typing import Iterable

import numpy as np
from shapely import GeometryCollection
from shapely import Polygon
from shapely import STRtree
from shapely import equals
from shapely import get_parts
from shapely import make_valid
from shapely import remove_repeated_points
from shapely import to_geojson
from shapely import unary_union
from shapely.geometry.base import BaseGeometry

from brdr.constants import BUFFER_MULTIPLICATION_FACTOR
from brdr.constants import CORR_DISTANCE
from brdr.constants import DEFAULT_CRS
from brdr.constants import GRB_VERSION_DATE
from brdr.constants import THRESHOLD_CIRCLE_RATIO
from brdr.enums import GRBType
from brdr.enums import OpenbaarDomeinStrategy
from brdr.geometry_utils import buffer_neg
from brdr.geometry_utils import buffer_neg_pos
from brdr.geometry_utils import buffer_pos
from brdr.geometry_utils import calculate_geom_by_intersection_and_reference
from brdr.geometry_utils import fill_and_remove_gaps
from brdr.geometry_utils import safe_difference
from brdr.geometry_utils import safe_intersection
from brdr.geometry_utils import safe_symmetric_difference
from brdr.geometry_utils import safe_union
from brdr.grb import GRBActualLoader
from brdr.loader import DictLoader
from brdr.loader import GeoJsonFileLoader
from brdr.loader import GeoJsonLoader
from brdr.loader import GeoJsonUrlLoader
from brdr.loader import Loader
from brdr.logger import Logger
from brdr.typings import ProcessResult
from brdr.utils import diffs_from_dict_series, multipolygons_to_singles
from brdr.utils import geojson_from_dict
from brdr.utils import get_breakpoints_zerostreak
from brdr.utils import get_series_geojson_dict
from brdr.utils import merge_dict
from brdr.utils import merge_dict_series
from brdr.utils import merge_process_results
from brdr.utils import write_geojson

date_format = "%Y-%m-%d"

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(message)s", datefmt="%d-%b-%y %H:%M:%S"
)


###################


class Aligner:
    """
    This class is used to compare the thematic data with the reference data.
    The reference data can be loaded in different ways, for example by using the GRB
    data.
    The thematic data can be loaded by using a geojson file.
    The class can be used to compare the thematic data with the reference data.
    """

    def __init__(
        self,
        *,
        feedback=None,
        relevant_distance=1,
        threshold_overlap_percentage=50,
        od_strategy=OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE,
        crs=DEFAULT_CRS,
        area_limit=None,
    ):
        """
        Initializes the Aligner object

        Args:
            feedback (object, optional): Feedback object that can be added to show
                feedback in QGIS. Defaults to None.
            relevant_distance (int, optional): The relevant distance (in meters) for
                processing. Defaults to 1.
            od_strategy (int, optional): The strategy to determine how to handle
                information outside the reference polygons (Openbaar Domein)
                (default 1: SNAP_SINGLE_SIDE)
            threshold_overlap_percentage (int, optional): Threshold (%) to determine
                from which overlapping-percentage a reference-polygon has to be included
                when there aren't relevant intersections or relevant differences
                (default 50%)
            od_strategy (int, optional): Determines how the algorithm deals with parts
                of the geometry that are not on the
                reference (default 1: SNAP_SINGLE_SIDE)
            crs (str, optional): Coordinate Reference System (CRS) of the data.
                (default EPSG:31370)
            area_limit (int, optional): Maximum area for processing. (default 100000)


        """
        self.logger = Logger(feedback)
        self.relevant_distance = relevant_distance
        self.od_strategy = od_strategy
        self.threshold_overlap_percentage = threshold_overlap_percentage
        self.area_limit = area_limit

        # PROCESSING DEFAULTS
        # thematic
        # name of the identifier-field of the thematic data (id has to be unique)
        self.name_thematic_id = "theme_identifier"
        # dictionary to store all thematic geometries to handle
        self.dict_thematic: dict[str, BaseGeometry] = {}
        # dictionary to store properties of the reference-features (optional)
        self.dict_thematic_properties: dict[str, dict] = {}
        # Dict to store source-information of the thematic dictionary
        self.dict_thematic_source: dict[str, str] = {}
        # dictionary to store all unioned thematic geometries
        self.thematic_union = None

        # reference

        # name of the identifier-field of the reference data (id has to be unique,f.e
        # CAPAKEY for GRB-parcels)
        self.name_reference_id = "ref_identifier"
        # dictionary to store all reference geometries
        self.dict_reference: dict[str, BaseGeometry] = {}
        # dictionary to store properties of the reference-features (optional)
        self.dict_reference_properties: dict[str, dict] = {}
        # Dict to store source-information of the reference dictionary
        self.dict_reference_source: dict[str, str] = {}
        # to save a unioned geometry of all reference polygons; needed for calculation
        # in most OD-strategies
        self.reference_union = None

        # results

        # output-dictionaries (when processing dict_thematic)
        self.dict_result: dict[str, ProcessResult] = {}
        # dictionary with the 'predicted' results, grouped by relevant distance
        self.dict_predicted = dict[float, dict[str, ProcessResult]]

        # Coordinate reference system
        # thematic geometries and reference geometries are assumed to be in the same CRS
        # before loading into the Aligner. No CRS-transformation will be performed.
        # When loading data, CRS is expected to be a projected CRS with units in 'meter
        # (m)'.
        # Default EPSG:31370 (Lambert72), alternative: EPSG:3812 (Lambert2008)
        self.CRS = crs
        # this parameter is used to treat multipolygon as single polygons. So polygons
        # with ID splitter are separately evaluated and merged on result.
        self.multi_as_single_modus = True
        self.logger.feedback_info("Aligner initialized")

    def buffer_distance(self):
        return self.relevant_distance / 2

    def process_geometry(
        self,
        input_geometry: BaseGeometry,
        relevant_distance=1,
        od_strategy=OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE,
        threshold_overlap_percentage=50,
    ) -> ProcessResult:
        """
        method to align a geometry to the reference layer

        Args:
            input_geometry (BaseGeometry): The input geometric object.
            relevant_distance
            od_strategy
            threshold_overlap_percentage (int): The buffer distance (positive
                or negative).

        Returns:
            ProcessResult : A dict containing the resulting geometries:

            *   result (BaseGeometry): The resulting output geometry
            *   result_diff (BaseGeometry): The resulting difference output geometry
            *   result_diff_plus (BaseGeometry): The resulting positive difference
                output geometry
            *   result_diff_min (BaseGeometry): The resulting negative difference output
                geometry
            *   relevant_intersection (BaseGeometry): The relevant_intersection
            *   relevant_difference (BaseGeometry): The relevant_difference
        Notes:
            -
        Example:
        """
        if self.area_limit and input_geometry.area > self.area_limit:
            message = "The input geometry is too large to process."
            raise ValueError(message)

        self.logger.feedback_debug("process geometry")
        self.relevant_distance = relevant_distance
        self.od_strategy = od_strategy
        self.threshold_overlap_percentage = threshold_overlap_percentage

        # combine all parts of the input geometry to one polygon
        input_geometry = unary_union(get_parts(input_geometry))

        # array with all relevant parts of a thematic geometry; initial empty Polygon
        (
            geometry,
            preresult,
            relevant_intersection_array,
            relevant_diff_array,
        ) = self._calculate_intersection_between_geometry_and_od(input_geometry)
        # get a list of all ref_ids that are intersecting the thematic geometry
        ref_intersections = self.reference_items.take(
            self.reference_tree.query(geometry)
        ).tolist()
        for key_ref in ref_intersections:
            geom_reference = self.dict_reference[key_ref]
            geom_intersection = safe_intersection(geometry, geom_reference)
            if geom_intersection.is_empty or geom_intersection is None:
                continue
            self.logger.feedback_debug("calculate intersection")
            (
                geom,
                relevant_intersection,
                relevant_diff,
            ) = calculate_geom_by_intersection_and_reference(
                geom_intersection,
                geom_reference,
                False,
                self.relevant_distance / 2,
                self.threshold_overlap_percentage,
            )
            self.logger.feedback_debug("intersection calculated")
            preresult = self.add_multi_polygons_from_geom_to_array(geom, preresult)
            relevant_intersection_array = self.add_multi_polygons_from_geom_to_array(
                relevant_intersection, relevant_intersection_array
            )
            relevant_diff_array = self.add_multi_polygons_from_geom_to_array(
                relevant_diff, relevant_diff_array
            )
        # UNION INTERMEDIATE LAYERS
        relevant_intersection = unary_union(relevant_intersection_array)
        if relevant_intersection is None or relevant_intersection.is_empty:
            relevant_intersection = Polygon()
        relevant_diff = unary_union(relevant_diff_array)
        if relevant_diff is None or relevant_diff.is_empty:
            relevant_diff = Polygon()

        # POSTPROCESSING
        result_dict = self._postprocess_preresult(preresult, geometry)

        result_dict["result_relevant_intersection"] = relevant_intersection
        result_dict["result_relevant_diff"] = relevant_diff

        # make a unary union for each key value in the result dict
        for key in ProcessResult.__annotations__:
            geometry = result_dict.get(key, Polygon())  # noqa
            if not geometry.is_empty:
                geometry = unary_union(geometry)
            result_dict[key] = geometry  # noqa

        return result_dict

    def process_dict_thematic(
        self,
        relevant_distance=1,
        od_strategy=OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE,
        threshold_overlap_percentage=50,
    ) -> dict[str, ProcessResult]:
        """
        Aligns a thematic dictionary of geometries to the reference layer based on
        specified parameters. - method to align a thematic dictionary to the reference
        layer

        Args:
            relevant_distance (float, optional): The relevant distance (in meters) for
                processing. Defaults to 1.
            od_strategy (int, optional): The strategy for overlap detection.
                Defaults to 1.
            threshold_overlap_percentage (float, optional): The threshold percentage for
                considering full overlap. Defaults to 50.

        Returns:
            dict: A dict containing processed data for each thematic key:
                - result: Aligned thematic data.
                - result_diff: global differences between thematic data and reference
                  data.
                - result_diff_plus: Positive differences.
                - result_diff_min: Negative differences.
                - relevant_intersection: relevant intersections.
                - relevant_diff: relevant differences.

        """
        dict_result = {}
        dict_thematic = self.dict_thematic
        if self.multi_as_single_modus:
            dict_thematic = multipolygons_to_singles(dict_thematic)
        for key in dict_thematic:
            self.logger.feedback_info("thematic id to process: " + str(key))
            try:
                dict_result[key] = self.process_geometry(
                    dict_thematic[key],
                    relevant_distance,
                    od_strategy,
                    threshold_overlap_percentage,
                )
            except ValueError as e:
                self.logger.feedback_warning(str(e))
        if self.multi_as_single_modus:
            dict_result = merge_process_results(dict_result)
        self.dict_result = dict_result
        return self.dict_result

    def predictor(
        self,
        relevant_distances=np.arange(0, 300, 10, dtype=int) / 100,
        od_strategy=OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE,
        threshold_overlap_percentage=50,
    ):
        """
        Predicts the 'most interesting' relevant distances for changes in thematic
        elements based on a distance series.

        This function analyzes a set of thematic geometries (`self.dict_thematic`) to
        identify potentially interesting distances where changes occur. It performs
        the following steps:

        1. **Process Distance Series:**
            - Calculates a series of results for different distances specified by
              `relevant_distances`.
            - This calculation might involve functions like `self.process_series`
              (implementation details likely depend on your specific code).

        2. **Calculate Difference Metrics:**
            - Analyzes the results from the distance series to compute difference
              metrics between thematic elements at each distance (using
              `diffs_from_dict_series`).

        3. **Identify Breakpoints and Zero-Streaks:**
            - For each thematic geometry, it identifies potential "breakpoints" where
              the difference metric changes sign (from positive to negative or vice
              versa).
            - It also identifies "zero-streaks" which are consecutive distances with a
              difference metric close to zero (potentially indicating minimal change).

        4. **Predict Interesting Distances:**
            - The function considers distances corresponding to breakpoints and
              zero-streaks as potentially interesting for further analysis.
            - These distances are stored in a dictionary (`dict_predicted`) with the
              thematic element key as the outer key.
            - Additionally, the corresponding results from the distance series for
              those distances are included.

        5. **Filter Results:**
            - The function might further filter the predicted results for each thematic
              element based on the element key (using `filter_resulting_series_by_key`).

        Args:
            relevant_distances (np.ndarray, optional): A NumPy array of distances to
              be analyzed. Defaults to np.arange(0.1, 5.05, 0.1).
            od_strategy (OpenbaarDomeinStrategy, optional): A strategy for handling
              open data in the processing (implementation specific). Defaults to
             OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE.
            threshold_overlap_percentage (int, optional): A percentage threshold for
              considering full overlap in the processing (implementation specific).
             Defaults to 50.

        Returns:
            dict: A dictionary containing predicted interesting distances for each
            thematic element.
                - Keys: Thematic element identifiers from `self.dict_thematic`.
                - Values: Dictionaries with the following structure for each
                   thematic element:
                    - Keys: Distances identified as interesting (breakpoints or
                    zero-streaks).
                    - Values: dicts containing results (likely specific to
                    your implementation) from the distance series for the
                    corresponding distance.

        Logs:
            - Debug logs the thematic element key being processed.
        """
        dict_predicted = defaultdict(dict)
        dict_series = self.process_series(
            relevant_distances=relevant_distances,
            od_strategy=od_strategy,
            threshold_overlap_percentage=threshold_overlap_percentage,
        )
        dict_thematic = self.dict_thematic
        if self.multi_as_single_modus:
            dict_series = merge_dict_series(dict_series)
            dict_thematic = merge_dict(self.dict_thematic)

        diffs_dict = diffs_from_dict_series(dict_series, dict_thematic)

        for theme_id, diffs in diffs_dict.items():
            if len(diffs) != len(relevant_distances):
                logging.warning(
                    f"Number of computed diffs for thematic element {theme_id} does "
                    f"not match the number of relevant distances."
                )
                continue
            diff_values = list(diffs.values())
            breakpoints, zero_streaks = get_breakpoints_zerostreak(
                relevant_distances, diff_values
            )
            logging.debug(str(theme_id))
            if len(zero_streaks) == 0:
                dict_predicted[relevant_distances[0]][theme_id] = dict_series[
                    relevant_distances[0]
                ][theme_id]
                logging.info("No zero-streaks found for: " + str(theme_id))
            for zs in zero_streaks:
                dict_predicted[zs[0]][theme_id] = dict_series[zs[0]][theme_id]

        self.dict_predicted = dict_predicted

        return (
            dict_series,
            dict_predicted,
            diffs_dict,
        )

    def process_series(
        self,
        relevant_distances: Iterable[float],
        od_strategy=OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE,
        threshold_overlap_percentage=50,
    ) -> dict[float, dict[str, ProcessResult]]:
        """
        Calculates the resulting dictionaries for thematic data based on a series of
            relevant distances.

        Args:
            relevant_distances (Iterable[float]): A series of relevant distances
                (in meters) to process
            od_strategy (int, optional): The strategy for overlap detection.
                Defaults to 1.
            threshold_overlap_percentage (float, optional): The threshold percentage for
                considering full overlap. Defaults to 50.

        Returns:
            dict: A dictionary containing the resulting dictionaries for a series of
                relevant distances:

                {
                    'relevant_distance_1': {theme_id_1: (ProcessResult), theme_id_2:
                        (ProcessResult), ...},
                    'relevant_distance_2': {theme_id_1: (ProcessResult), theme_id_2:
                        (ProcessResult), ...},
                    ...
                }
        """
        self.logger.feedback_debug("Process series" + str(relevant_distances))
        self.od_strategy = od_strategy
        self.threshold_overlap_percentage = threshold_overlap_percentage
        dict_series = {}
        for s in relevant_distances:
            self.logger.feedback_info(
                "Processing series - relevant_distance (m):"
                + str(s)
                + " with ODStrategy "
                + str(self.od_strategy)
            )
            dict_series[s] = self.process_dict_thematic(s, od_strategy)
        self.logger.feedback_info(
            "End of processing series: " + str(relevant_distances)
        )
        return dict_series

    def get_formula(self, geometry: BaseGeometry, with_geom=False):
        """
        Calculates formula-related information based on the input geometry.

        Args:
            geometry (shapely.geometry object): The input geometry.
            with_geom (bool, optional): Whether to include geometry information in the
                output. Defaults to False.

        Returns:
            dict: A dictionary containing formula-related data:

                -   'full': True if the intersection is the same as the reference
                    geometry, else False.
                -   'area': Area of the intersection or reference geometry.
                -   'percentage': Percentage of intersection area relative to the
                    reference geometry.
                -   'geometry': GeoJSON representation of the intersection (if
                    with_geom is True).
        """
        dict_formula = {
            "alignment_date": datetime.now().strftime(date_format),
            "reference_source": self.dict_reference_source,
            "full": True,
            "reference_features": {},
            "reference_od": None,
            "last_version_date": None,
        }

        full_total = True
        last_version_date = None

        ref_intersections = self.reference_items.take(
            self.reference_tree.query(geometry)
        ).tolist()
        intersected = []
        for key_ref in ref_intersections:
            geom = None
            geom_reference = self.dict_reference[key_ref]
            geom_intersection = make_valid(safe_intersection(geometry, geom_reference))
            if geom_intersection.is_empty or geom_intersection is None:
                continue
            intersected.append(geom_intersection)
            # Add a last_version_date if available in properties
            if (
                key_ref in self.dict_reference_properties
                and GRB_VERSION_DATE in self.dict_reference_properties[key_ref]
            ):
                version_date = datetime.strptime(
                    self.dict_reference_properties[key_ref][GRB_VERSION_DATE][:10],
                    date_format,
                ).date()
                if last_version_date is None and version_date is not None:
                    last_version_date = version_date
                if version_date is not None and version_date > last_version_date:
                    last_version_date = version_date
            if equals(geom_intersection, geom_reference):
                full = True
                area = round(geom_reference.area, 2)
                perc = 100
                if with_geom:
                    geom = to_geojson(geom_reference)
            else:
                perc = round(geom_intersection.area * 100 / geom_reference.area, 2)
                if perc < 0.01:
                    continue
                elif perc > 99.99:
                    full = True
                    area = round(geom_reference.area, 2)
                    perc = 100
                    if with_geom:
                        geom = to_geojson(geom_reference)
                else:
                    full = False
                    full_total = False
                    area = round(geom_intersection.area, 2)
                    if with_geom:
                        geom = to_geojson(geom_intersection)

            dict_formula["reference_features"][key_ref] = {
                "full": full,
                "area": area,
                "percentage": perc,
                "geometry": geom,
            }
        dict_formula["full"] = full_total
        if last_version_date is not None:
            dict_formula["versiondate"] = last_version_date.strftime(date_format)
        geom_od = safe_difference(geometry, make_valid(unary_union(intersected)))
        if geom_od is not None:
            area_od = round(geom_od.area, 2)
            if area_od > 0:
                dict_formula["reference_od"] = {
                    "area": area_od,
                    "geometry": to_geojson(geom_od),
                }
        self.logger.feedback_debug(str(dict_formula))
        return dict_formula

    def get_results_as_dict(self):
        """
        get a dict of the results
        """
        if self.multi_as_single_modus:
            return merge_process_results(self.dict_result)
        return self.dict_result

    def get_results_as_geojson(self, formula=False):
        """
        convert the results to geojson feature collections

        Args:
            formula (bool, optional): Whether to include formula-related information
                in the output. Defaults to False.
        """
        results_dict = self.dict_result
        if self.multi_as_single_modus:
            results_dict = merge_process_results(results_dict)

        return self.get_predictions_as_geojson(
            formula,
            {self.relevant_distance: results_dict},
        )

    def get_predictions_as_geojson(self, formula=False, series_dict=None):
        """
        get a dictionary containing of the resulting geometries as geojson, based on the
            'predicted' relevant distances.
        Optional: The descriptive formula is added as an attribute to the result"""

        series_dict = series_dict or self.dict_predicted
        prop_dictionary = defaultdict(dict)

        for relevant_distance, results_dict in series_dict.items():
            for theme_id, process_results in results_dict.items():
                if formula:
                    result = process_results["result"]
                    formula = self.get_formula(result)
                    prop_dictionary[relevant_distance][theme_id] = {
                        "formula": json.dumps(formula)
                    }

        return get_series_geojson_dict(
            series_dict,
            crs=self.CRS,
            id_field=self.name_thematic_id,
            series_prop_dict=prop_dictionary,
        )

    def get_reference_as_geojson(self):
        """
        get a geojson of the reference polygons
        """
        return geojson_from_dict(
            self.dict_reference, self.CRS, self.name_reference_id, geom_attributes=False
        )

    def export_results(self, path, formula=True):
        """
        Exports analysis results as GeoJSON files.

        This function exports 6 GeoJSON files containing the analysis results to the
            specified `path`.

        Args:
            path (str): The path to the directory where the GeoJSON files will be saved.
            formula (bool, optional): Whether to include formula-related information
                in the output. Defaults to True.

        Details of exported files:
            - result.geojson: Contains the original thematic data from `
              self.dict_result`.
            - result_diff.geojson: Contains the difference between the original
              and predicted data from `self.dict_result_diff`.
            - result_diff_plus.geojson: Contains results for areas that are
              added (increased area).
            - result_diff_min.geojson: Contains results for areas that are
              removed (decreased area).
            - result_relevant_intersection.geojson: Contains the areas with
              relevant intersection that has to be included in the result.
            - result_relevant_difference.geojson: Contains the areas with
              relevant difference that has to be excluded from the result.
        """
        fcs = self.get_results_as_geojson(formula)
        for name, fc in fcs.items():
            write_geojson(os.path.join(path, name + ".geojson"), fc)

    def _prepare_reference_data(self):
        """
        Prepares reference data for spatial queries and analysis.

        It performs the following tasks:

        1. **Optimizes spatial queries:**
            - Creates a Spatial Relationship Tree (STRtree) using `STRtree` for
              efficient spatial queries against the reference data in
              `self.dict_reference`.
            - Converts the dictionary keys (reference identifiers) to a NumPy array
              for potential performance benefits in future operations.

        2. **Clears reference union:**
            - Sets `self.reference_union` to `None`. This variable stores the combined
              geometry of all reference data, and it's cleared here to indicate that
              it needs to be recalculated if requested later.

        Returns:
            None
        """
        # create an SRTree for performance optimisation
        self.logger.feedback_info(
            "length of reference_dict: " + str(len(self.dict_reference))
        )
        self.reference_tree = STRtree(list(self.dict_reference.values()))
        self.reference_items = np.array(list(self.dict_reference.keys()))
        # clear the reference_union, so it will be recalculated on request when needed
        self.reference_union = None
        return

    def _calculate_intersection_between_geometry_and_od(self, geometry):
        # Calculate the intersection between thematic and Openbaar Domein
        relevant_intersection_array = []
        relevant_difference_array = []
        geom_thematic_od = Polygon()

        if self.od_strategy == OpenbaarDomeinStrategy.EXCLUDE:
            # Completely exclude everything that is not on the reference layer
            self.logger.feedback_debug("OD-strategy EXCLUDE")
            # Remove from the thematic layer all parts that are not on the reference
            # layer
            # !!this strategy adapts the input-geometry!!
            geometry = safe_intersection(geometry, self._get_reference_union())
        elif self.od_strategy == OpenbaarDomeinStrategy.AS_IS:
            # All parts that are not covered by the reference layer are added to the
            #         resulting geometry AS IS
            self.logger.feedback_debug("OD-strategy AS IS")
            # all OD-parts wil be added AS IS
            geom_thematic_od = safe_difference(geometry, self._get_reference_union())
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE:
            # Everything that falls within the relevant distance over the
            #  plot boundary is snapped to the plot.
            #  Only the inner-reference-boundaries are used.
            #  The outer-reference-boundaries are not used.
            self.logger.feedback_debug("OD-strategy SNAP_SINGLE_SIDE")
            # geom of OD
            geom_od = safe_difference(geometry, self._get_reference_union())
            # only the relevant parts of OD
            geom_od_neg_pos = buffer_neg_pos(geom_od, self.buffer_distance())
            # geom_thematic_od = safe_intersection(geom_od_neg_pos,geom_od)# resulting
            # thematic OD
            geom_od_neg_pos_buffered = buffer_pos(
                geom_od_neg_pos, self.buffer_distance()
            )  # include parts
            geom_thematic_od = safe_intersection(
                geom_od_neg_pos_buffered, geom_od
            )  # resulting thematic OD
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_ALL_SIDE:
            # Everything that falls within the relevant distance over
            # the plot boundary is snapped to the plot.
            #  Inner-reference-boundaries and outer-reference-boundaries are used.
            self.logger.feedback_debug("OD-strategy SNAP BOTH SIDED")
            (
                geom_thematic_od,
                relevant_difference_array,
                relevant_intersection_array,
            ) = self._od_snap_all_side(geometry)
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_FULL_AREA_SINGLE_SIDE:
            # Strategy useful for bigger areas.
            # integrates the entire inner area of the input geometry,
            # so Openbaar Domein of the inner area is included in the result
            # Combines SNAP_SINGLE_SIDE with the inner area
            self.logger.feedback_debug(
                "OD-strategy Full-area-variant of OD-SNAP_SINGLE_SIDE"
            )
            geom_thematic_od = self._od_full_area(geometry)
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE:
            # Strategy useful for bigger areas.
            # integrates the entire inner area of the input geometry,
            # so Openbaar Domein of the inner area is included in the result
            # Combines SNAP_ALL_SIDE with the inner area
            self.logger.feedback_debug(
                "OD-strategy Full-area-variant of OD-SNAP_ALL_SIDE"
            )
            # first part is a copy of OD_ALL_SIDE
            (
                geom_thematic_od,
                relevant_difference_array,
                relevant_intersection_array,
            ) = self._od_snap_all_side(geometry)
            # This part is a copy of SNAP_FULL_AREA_SINGLE_SIDE
            geom_theme_od_min_clipped_plus_buffered_clipped = self._od_full_area(
                geometry
            )
            # UNION the calculation of  OD-SNAP_ALL_SIDE with FULL AREA of
            # OD-SNAP_FULL_AREA_SINGLE_SIDE
            geom_thematic_od = safe_union(
                geom_theme_od_min_clipped_plus_buffered_clipped, geom_thematic_od
            )

        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE_VARIANT_1:
            # OD-strategy SNAP_SINGLE_SIDE - variant 1:
            # Everything that falls within the relevant distance over the
            #  plot boundary is snapped to the plot.
            #  Only the inner-reference-boundaries are used.
            #  The outer-reference-boundaries are not used.
            self.logger.feedback_debug("OD-strategy SNAP_SINGLE_SIDE - variant 1")
            # geom of OD
            geom_od = safe_difference(geometry, self._get_reference_union())
            # only the relevant parts of OD
            geom_od_neg_pos = buffer_neg_pos(geom_od, self.buffer_distance())
            # resulting thematic OD
            geom_thematic_od = safe_intersection(geom_od_neg_pos, geom_od)
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE_VARIANT_2:
            # OD-strategy SNAP_SINGLE_SIDE - variant 2:
            # Everything that falls within the relevant distance over the
            #  plot boundary is snapped to the plot.
            #  Only the inner-reference-boundaries are used.
            #  The outer-reference-boundaries are not used.
            self.logger.feedback_debug("OD-strategy SNAP_SINGLE_SIDE - variant 2")
            geom_thematic_od = Polygon()
            pass

        # ADD THEMATIC_OD
        preresult = self.add_multi_polygons_from_geom_to_array(geom_thematic_od, [])
        return (
            geometry,
            preresult,
            relevant_intersection_array,
            relevant_difference_array,
        )

    def _od_full_area(self, geometry):
        geom_theme_od = safe_difference(geometry, self._get_reference_union())
        geom_theme_min_buffered = buffer_neg(
            buffer_pos(
                buffer_neg(geometry, self.relevant_distance),
                self.buffer_distance(),
            ),
            self.buffer_distance(),
        )
        geom_theme_od_clipped_min_buffered = safe_intersection(
            geom_theme_min_buffered, geom_theme_od
        )
        geom_theme_od_min_clipped_plus_buffered = buffer_pos(
            geom_theme_od_clipped_min_buffered, self.relevant_distance
        )
        geom_theme_od_min_clipped_plus_buffered_clipped = safe_intersection(
            geom_theme_od_min_clipped_plus_buffered, geom_theme_od
        )
        geom_thematic_od = geom_theme_od_min_clipped_plus_buffered_clipped
        return geom_thematic_od

    def _od_snap_all_side(self, geometry):
        relevant_difference_array = []
        relevant_intersection_array = []
        geom_thematic_buffered = make_valid(
            buffer_pos(geometry, BUFFER_MULTIPLICATION_FACTOR * self.relevant_distance)
        )
        clip_ref_thematic_buffered = safe_intersection(
            self._get_reference_union(), geom_thematic_buffered
        )
        geom_reference = safe_difference(
            geom_thematic_buffered, clip_ref_thematic_buffered
        )  # Both OD-parts are SNAPPED added
        geom_thematic = geometry
        if geom_reference.is_empty or geom_reference is None:
            geom_thematic_od = geom_reference
        else:
            geom_intersection = safe_intersection(geom_reference, geom_thematic)
            (
                geom_thematic_od,
                geom_relevant_intersection,
                geom_relevant_diff,
            ) = calculate_geom_by_intersection_and_reference(
                geom_intersection,
                geom_reference,
                True,
                self.relevant_distance / 2,
                self.threshold_overlap_percentage,
            )
            relevant_intersection_array = self.add_multi_polygons_from_geom_to_array(
                geom_relevant_intersection, []
            )
            relevant_difference_array = self.add_multi_polygons_from_geom_to_array(
                geom_relevant_diff, []
            )
        return geom_thematic_od, relevant_difference_array, relevant_intersection_array

    # def _snap_geom_to_reference(self, geom_input, geom_reference, relevant_distance):
    # """
    # This feature does not work correctly with Shapely. This avoids polygons collapse
    # if everything is taken together, which we do in some cases effectively want.
    # """
    # return snap(geom_input, geom_reference, relevant_distance)

    def _get_reference_union(self):
        if self.reference_union is None:
            self.reference_union = make_valid(
                unary_union(list(self.dict_reference.values()))
            )
        return self.reference_union

    def get_thematic_union(self):
        if self.thematic_union is None:
            self.thematic_union = make_valid(
                unary_union(list(self.dict_thematic.values()))
            )
        return self.thematic_union

    def _postprocess_preresult(self, preresult, geom_thematic) -> ProcessResult:
        """
        Postprocess the preresult with the following actions to create the final result
        *Corrections for areas that differ more than the relevant distance
        *slivers
        *Inner holes (donuts) /multipolygons
        *validity
        *Circles (Polspy-popper)
        *Null/Empty-values

        Args:
            preresult (list): An existing list with all the elements of the preresult
            geom_thematic (BaseGeometry): The input geometry

        Returns:
            ProcessResult: A dictionary containing the resulting output geometries:

            *   result (BaseGeometry): The resulting output geometry
            *   result_diff (BaseGeometry): The resulting difference output geometry
            *   result_diff_plus (BaseGeometry): The resulting positive difference
                output geometry
            *   result_diff_min (BaseGeometry): The resulting negative difference output
                geometry
        """
        # Process array
        result = []
        geom_preresult = make_valid(unary_union(preresult))
        geom_thematic = make_valid(geom_thematic)

        if not (geom_thematic is None or geom_thematic.is_empty):
            # Correction for circles
            # calculate ratio to see if it is a circle, and keep the original geometry
            #  if a circle: (Polsby-popper score)
            if (
                4 * pi * (geom_thematic.area / (geom_thematic.length**2))
                > THRESHOLD_CIRCLE_RATIO
            ):
                self.logger.feedback_warning(
                    "Circle: -->resulting geometry = original geometry"
                )
                return {"result": geom_thematic}

            # Correction for unchanged geometries
            if geom_preresult == geom_thematic:
                return {"result": geom_thematic}

        # Corrections for areas that differ more than the relevant distance
        geom_thematic_dissolved = buffer_pos(
            buffer_neg(
                buffer_pos(geom_preresult, CORR_DISTANCE),
                2 * CORR_DISTANCE,
            ),
            CORR_DISTANCE,
        )
        # geom_symdiff = self._safe_symmetric_difference(geom_thematic,
        # geom_thematic_dissolved)
        geom_diff_add = safe_difference(geom_thematic, geom_thematic_dissolved)
        geom_diff_delete = safe_difference(geom_thematic_dissolved, geom_thematic)
        geom_diff_removed = safe_difference(
            geom_thematic_dissolved,
            safe_intersection(
                geom_diff_delete,
                buffer_neg_pos(geom_diff_delete, self.buffer_distance()),
            ),
        )
        geom_diff_removed_added = safe_union(
            geom_diff_removed,
            safe_intersection(
                geom_diff_add,
                buffer_neg_pos(geom_diff_add, self.buffer_distance()),
            ),
        )
        geom_thematic_preresult = buffer_pos(
            buffer_neg(
                buffer_pos(geom_diff_removed_added, CORR_DISTANCE),
                2 * CORR_DISTANCE,
            ),
            CORR_DISTANCE,
        )
        # Correction for Inner holes(donuts) / multipolygons
        # fill and remove gaps
        geom_thematic_cleaned_holes = fill_and_remove_gaps(
            geom_thematic_preresult, self.buffer_distance()
        )
        geom_thematic_result = buffer_pos(
            buffer_neg(
                buffer_pos(geom_thematic_cleaned_holes, CORR_DISTANCE),
                2 * CORR_DISTANCE,
            ),
            CORR_DISTANCE,
        )
        geom_thematic_result = make_valid(remove_repeated_points(geom_thematic_result))

        # Correction for empty preresults
        if geom_thematic_result.is_empty or geom_thematic_result is None:
            self.logger.feedback_warning(
                "Empty result: -->resulting geometry = empty geometry"
            )
            # geom_thematic_result = geom_thematic
            geom_thematic_result = Polygon()

        # group all initial multipolygons into a new resulting dictionary
        result.append(geom_thematic_result)

        # create all resulting geometries
        geom_thematic_result = make_valid(unary_union(result))

        # negative and positive buffer is added to the difference-calculations, to
        # remove 'very small' differences (smaller than the correction distance)
        geom_result_diff = buffer_pos(
            buffer_neg(
                safe_symmetric_difference(geom_thematic_result, geom_thematic),
                CORR_DISTANCE,
            ),
            CORR_DISTANCE,
        )
        geom_result_diff_plus = buffer_pos(
            buffer_neg(
                safe_difference(geom_thematic_result, geom_thematic),
                CORR_DISTANCE,
            ),
            CORR_DISTANCE,
        )
        geom_result_diff_min = buffer_pos(
            buffer_neg(
                safe_difference(geom_thematic, geom_thematic_result),
                CORR_DISTANCE,
            ),
            CORR_DISTANCE,
        )
        # geom_result_diff_plus = safe_difference(geom_thematic_result, geom_thematic)
        # geom_result_diff_min = safe_difference(geom_thematic, geom_thematic_result)

        return {
            "result": geom_thematic_result,
            "result_diff": geom_result_diff,
            "result_diff_plus": geom_result_diff_plus,
            "result_diff_min": geom_result_diff_min,
        }

    @staticmethod
    def add_multi_polygons_from_geom_to_array(geom: BaseGeometry, array):
        """
        Append valid polygons and multipolygons extracted from a given geometry to an
        existing array.

        Args:
            geom (BaseGeometry): The input geometry to process.
            array (list): An existing list to store valid polygons and multipolygons.

        Returns:
            list: A list containing valid polygons and multipolygons extracted from the
                input geometry.
        """
        if geom.is_empty or geom is None:
            # If the input geometry is empty or None, do nothing.
            pass
        else:
            # Create a GeometryCollection from the input geometry.
            geometry_collection = GeometryCollection(geom)  # noqa
            for g in geometry_collection.geoms:
                # Ensure each sub-geometry is valid.
                g = make_valid(g)
                if str(g.geom_type) in ["Polygon", "MultiPolygon"]:
                    # Append valid polygons and multipolygons to the array.
                    array.append(g)
        return array

    def load_reference_data(self, loader: Loader):
        (
            self.dict_reference,
            self.dict_reference_properties,
            self.dict_reference_source,
        ) = loader.load_data()
        self._prepare_reference_data()

    def load_thematic_data(self, loader: Loader):
        self.dict_thematic, self.dict_thematic_properties, self.dict_thematic_source = (
            loader.load_data()
        )

        # Deprecated loader methods

    def load_thematic_data_geojson(self, thematic_input, name_thematic_id):
        logging.warning("deprecated method, use load_thematic_data instead")
        loader = GeoJsonLoader(thematic_input, name_thematic_id)
        self.load_thematic_data(loader)

    def load_thematic_data_file(self, path_to_file, name_thematic_id):
        logging.warning("deprecated method, use load_thematic_data instead")
        loader = GeoJsonFileLoader(path_to_file, name_thematic_id)
        self.load_thematic_data(loader)

    def load_thematic_data_dict(self, dict_theme):
        logging.warning("deprecated method, use load_thematic_data instead")
        loader = DictLoader(dict_theme)
        self.load_thematic_data(loader)

    def load_thematic_data_url(self, url, name_thematic_id):
        logging.warning("deprecated method, use load_thematic_data instead")
        loader = GeoJsonUrlLoader(url, name_thematic_id)
        self.load_thematic_data(loader)

    def load_reference_data_dict(self, dict_ref):
        logging.warning("deprecated method, use load_reference_data instead")
        loader = DictLoader(dict_ref)
        self.load_reference_data(loader)

    def load_reference_data_geojson(self, reference_input, name_reference_id):
        logging.warning("deprecated method, use load_reference_data instead")
        loader = GeoJsonLoader(_input=reference_input, id_property=name_reference_id)
        self.load_reference_data(loader)

    def load_reference_data_file(self, path_to_file, name_reference_id):
        logging.warning("deprecated method, use load_reference_data instead")
        loader = GeoJsonFileLoader(path_to_file, name_reference_id)
        self.load_reference_data(loader)

    def load_reference_data_url(self, url, name_reference_id):
        logging.warning("deprecated method, use load_reference_data instead")
        loader = GeoJsonUrlLoader(url, name_reference_id)
        self.load_reference_data(loader)

    def load_reference_data_grb_actual(self, *, grb_type=GRBType.ADP, partition=1000):
        logging.warning("deprecated method, use load_reference_data instead")
        loader = GRBActualLoader(grb_type=grb_type, partition=partition, aligner=self)
        self.load_reference_data(loader)



from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.utils import get_oe_dict_by_ids
from brdr.utils import multipolygons_to_singles
from examples import print_formula
from examples import show_map

# example to Change a dictionary form multipolygon to single before executing the
# aligner. Can be used on the thematic dictionary as the reference dictionary


if __name__ == "__main__":
    # EXAMPLE for a thematic MultiPolygon
    dict_theme = get_oe_dict_by_ids([110082])

    # WITHOUT MULTI_TO_SINGLE
    # Initiate brdr
    aligner = Aligner()
    # Load thematic data & reference data
    # Get a specific feature of OE that exists out of a Multipolygon

    aligner.load_thematic_data_dict(dict_theme)
    aligner.load_reference_data_grb_actual(grb_type=GRBType.GBG, partition=1000)

    rel_dist = 2
    dict_results_by_distance = {rel_dist: aligner.process_dict_thematic(rel_dist, 4)}
    aligner.export_results("output/")
    show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)

    print_formula(dict_results_by_distance, aligner)

    # WITH MULTI_TO_SINGLE
    # Initiate brdr
    aligner = Aligner()
    # Load thematic data & reference data
    # Get a specific feature of OE that exists out of a Multipolygon
    dict_theme = multipolygons_to_singles(dict_theme)
    aligner.load_thematic_data_dict(dict_theme)
    aligner.load_reference_data_grb_actual(grb_type=GRBType.GBG, partition=1000)

    rel_dist = 5
    dict_results_by_distance = {rel_dist: aligner.process_dict_thematic(rel_dist, 4)}
    aligner.export_results("output/")
    show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)

    print_formula(dict_results_by_distance, aligner)



import json
import logging
from datetime import date
from datetime import datetime

from shapely import intersects
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry

from brdr.constants import DEFAULT_CRS
from brdr.constants import DOWNLOAD_LIMIT
from brdr.constants import GRB_BUILDING_ID
from brdr.constants import GRB_FEATURE_URL
from brdr.constants import GRB_FISCAL_PARCELS_URL
from brdr.constants import GRB_KNW_ID
from brdr.constants import GRB_PARCEL_ID
from brdr.constants import GRB_VERSION_DATE
from brdr.constants import MAX_REFERENCE_BUFFER
from brdr.enums import Evaluation
from brdr.enums import GRBType
from brdr.geometry_utils import buffer_pos
from brdr.geometry_utils import create_donut
from brdr.geometry_utils import features_by_geometric_operation
from brdr.geometry_utils import get_bbox
from brdr.loader import GeoJsonLoader
from brdr.utils import dict_series_by_keys
from brdr.utils import geojson_to_dicts
from brdr.utils import get_collection
from brdr.utils import get_collection_by_partition
from brdr.utils import merge_dict

log = logging.getLogger(__name__)
date_format = "%Y-%m-%d"


def is_grb_changed(
    geometry,
    grb_type=GRBType.ADP,
    date_start=date.today(),
    date_end=date.today(),
    border_distance=0,
    crs=DEFAULT_CRS,
):
    """
       checks if a geometry is possibly affected by changes in the reference layer
       during a specified timespan
    Args:
        geometry: Geometry to check on GRB-changes
        grb_type:  Type of GRB (parcels, buildings,artwork,...) to check
        date_start: Start of timespan to check if GRB-changes has occurred
        date_end: End of timespan to check if GRB-changes has occurred
        border_distance: Distance that can be used to only check the 'border' of
            the geometry, so 'big' geometries with internal parcel-updates are not
             affected
        (Default:0, indicating that the full geometry is checked fot GRB-changes)
        crs: Coordinate reference system to use

    Returns: Boolean, indicating if GRB is changed underneath the geometry

    """

    last_version_date = get_last_version_date(
        geometry, grb_type=grb_type, crs=crs, border_distance=border_distance
    )
    if last_version_date is None:
        return None
    if date_start <= last_version_date <= date_end:
        return True
    return False


def get_geoms_affected_by_grb_change(
    aligner,
    grb_type=GRBType.ADP,
    date_start=date.today(),
    date_end=date.today(),
    one_by_one=False,
    border_distance=0,
):
    """
    Get a dictionary of thematic geometries that are affected bij GRB-changes in a
    specific timespan

    Args:
        aligner: Aligner instance
        grb_type: Type of GRB: parcels, buildings,...
        date_start: start-date to check changes in GRB
        date_end: end-date to check changes in GRB
        one_by_one: parameter to choose the methodology to check changes:
            * True: Every thematic geometry is checked individually (loop)
            * False: All GRB-parcels intersecting the thematic dictionary is checked
                at once
        border_distance: Distance that can be used to only check the 'border' of  the
            geometry, so 'big' geometries with internal parcel-updates are not affected
            (Default:0, indicating that the full geometry is checked fot GRB-changes)
    Returns:
        dictionary of affected geometries

    """
    dict_thematic = aligner.dict_thematic
    if aligner.multi_as_single_modus:
        dict_thematic = merge_dict(dict_thematic)
    crs = aligner.CRS
    affected_dict: dict[str, BaseGeometry] = {}
    unchanged_dict: dict[str, BaseGeometry] = {}
    if border_distance > 0:
        for key in dict_thematic.keys():
            dict_thematic[key] = create_donut(dict_thematic[key], border_distance)
    if one_by_one:
        for key in dict_thematic:
            geom = dict_thematic[key]
            if is_grb_changed(geom, grb_type, date_start, date_end):
                affected_dict[key] = geom
            else:
                unchanged_dict[key] = geom
        return affected_dict, unchanged_dict
    else:
        # Temporal filter on VERDATUM
        geometry = aligner.get_thematic_union()
        coll_changed_grb, name_reference_id = get_collection_grb_actual(
            geometry,
            grb_type=grb_type,
            partition=1000,
            date_start=date_start,
            date_end=date_end,
            crs=crs,
        )
        dict_changed_grb, dict_changed_grb_properties = geojson_to_dicts(
            coll_changed_grb, name_reference_id
        )

        if len(dict_changed_grb) == 0:
            logging.info("No detected changes")
            return affected_dict, dict_thematic  # empty affected dict
        logging.info("Changed parcels in timespan: " + str(len(dict_changed_grb)))
        thematic_intersections = features_by_geometric_operation(
            list(dict_thematic.values()),
            list(dict_thematic.keys()),
            list(dict_changed_grb.values()),
            predicate="intersects",
        )
        logging.info("Number of filtered features: " + str(len(thematic_intersections)))
        for key, geom in dict_thematic.items():
            if key in thematic_intersections:
                affected_dict[key] = geom
            else:
                unchanged_dict[key] = geom
    return affected_dict, unchanged_dict


def get_last_version_date(
    geometry,
    grb_type=GRBType.ADP,
    crs=DEFAULT_CRS,
    limit=DOWNLOAD_LIMIT,
    border_distance=0,
):
    """
    Retrieves the date of the last version for a specific geographic area within
    GRB (parcels, buildings,...).

    This function queries the GRB-API to find the most recent version-date (=last
    update  of object) for reference data of the specified `grb_type` (e.g., ADP,
    GBG, KNW) within the boundary of the provided `geometry`.

    Args:
    geometry (BaseGeometry): A Shapely geometry representing the area of interest.
    grb_type (GRBType, optional): The type of GRB data to consider.
        Defaults to GRBType.ADP (administrative plots).

    Returns:
    str: The date of the last version for the specified GRB data type within the area,
         formatted as a string according to the GRB API response (usually YYYY-MM-DD).

    None: If no data is found for the given geometry and GRB type combination.
    """
    if border_distance > 0:
        geometry = create_donut(geometry, border_distance)
    bbox = get_bbox(geometry)
    actual_url = (
        GRB_FEATURE_URL + "/" + grb_type.upper() + "/items?"
        "limit=" + str(limit) + "&crs=" + crs + "&bbox-crs=" + crs + "&bbox=" + bbox
    )
    update_dates = []
    collection = get_collection(actual_url, limit)
    if "features" not in collection:
        return None
    for c in collection["features"]:
        if intersects(geometry, shape(c["geometry"])):
            versiondate = datetime.strptime(
                c["properties"][GRB_VERSION_DATE], date_format
            ).date()
            update_dates.append(versiondate)

    update_dates = sorted(update_dates, reverse=True)
    if len(update_dates) > 0:
        return update_dates[0]
    return None


def get_collection_grb_actual(
    geometry,
    grb_type=GRBType.ADP,
    partition=1000,
    limit=DOWNLOAD_LIMIT,
    crs=DEFAULT_CRS,
    date_start=None,
    date_end=None,
):

    url = (
        GRB_FEATURE_URL
        + "/"
        + grb_type.upper()
        + "/items?limit="
        + str(limit)
        + "&crs="
        + crs
    )
    if grb_type == GRBType.ADP:
        name_reference_id = GRB_PARCEL_ID
    elif grb_type == "gbg":
        name_reference_id = GRB_BUILDING_ID
    elif grb_type == GRBType.KNW:
        name_reference_id = GRB_KNW_ID
    else:
        logging.warning(
            f"type not implemented: {str(grb_type)} -->No reference-data loaded"
        )
        return {}, None

    versiondate_filter = ""
    versiondate_filter_start = ""
    versiondate_filter_end = ""
    if date_start is not None:
        versiondate_filter_start = (
            GRB_VERSION_DATE + ">" + date_start.strftime(date_format)
        )
        versiondate_filter = versiondate_filter_start
    if date_end is not None:
        versiondate_filter_end = GRB_VERSION_DATE + "<" + date_end.strftime(date_format)
        versiondate_filter = versiondate_filter_end
    if not (date_start is None and date_end is None):
        versiondate_filter = versiondate_filter_start + " AND " + versiondate_filter_end
    if versiondate_filter != "":
        url = url + "&filter=" + versiondate_filter + "&filter-lang=cql-text"

    collection = get_collection_by_partition(
        url, geometry=geometry, partition=partition, limit=limit, crs=crs
    )
    return collection, name_reference_id


def get_collection_grb_fiscal_parcels(
    geometry,
    year=str(datetime.now().year),
    partition=1000,
    limit=DOWNLOAD_LIMIT,
    crs=DEFAULT_CRS,
):
    url = (
        GRB_FISCAL_PARCELS_URL + "/Adpf" + year + "/items?"
        "limit=" + str(limit) + "&crs=" + crs
    )
    return get_collection_by_partition(
        url, geometry=geometry, partition=partition, limit=limit, crs=crs
    )


def evaluate(
    actual_aligner,
    dict_series,
    dict_predicted,
    thematic_dict_formula,
    threshold_area=5,
    threshold_percentage=1,
    dict_unchanged=None,
):
    """
    evaluate affected geometries and give attributes to evaluate and decide if new
    proposals can be used
    """
    if dict_unchanged is None:
        dict_unchanged = {}
    theme_ids = list(dict_series_by_keys(dict_series).keys())
    dict_evaluated_result = {}
    prop_dictionary = {}
    # Fill the dictionary-structure with empty values
    for dist in dict_series.keys():
        dict_evaluated_result[dist] = {}
        prop_dictionary[dist] = {}
        for theme_id in theme_ids:
            prop_dictionary[dist][theme_id] = {}
        for theme_id in dict_unchanged.keys():
            prop_dictionary[dist][theme_id] = {}

    equality = False
    for dist, dict_predicted_for_dist in dict_predicted.items():
        if equality:
            break
        for theme_id in dict_predicted_for_dist.keys():
            results = dict_series[dist][theme_id]
            actual_formula = actual_aligner.get_formula(results["result"])
            prop_dictionary[dist][theme_id]["formula"] = json.dumps(actual_formula)
            base_formula = None
            if theme_id in thematic_dict_formula:
                base_formula = thematic_dict_formula[theme_id]
            equality, _property = check_equality(
                base_formula,
                actual_formula,
                threshold_area,
                threshold_percentage,
            )
            if equality:
                dict_evaluated_result[dist][theme_id] = dict_predicted[dist][theme_id]
                prop_dictionary[dist][theme_id]["evaluation"] = _property
                break

    dict_predicted_keys = dict_series_by_keys(dict_predicted)
    evaluated_theme_ids = list(dict_series_by_keys(dict_evaluated_result).keys())
    # fill where no equality is found/ The biggest predicted distance is returned as
    # proposal
    for theme_id in theme_ids:
        if theme_id not in evaluated_theme_ids:
            if len(dict_predicted_keys[theme_id].keys()) == 0:
                result = dict_series[0][theme_id]
                dict_evaluated_result[0][theme_id] = result
                prop_dictionary[0][theme_id]["formula"] = json.dumps(
                    actual_aligner.get_formula(result["result"])
                )
                prop_dictionary[0][theme_id]["evaluation"] = Evaluation.NO_PREDICTION_5
                continue
            dist_predicted_max = max(list(dict_predicted_keys[theme_id].keys()))
            logging.debug(
                "max predicted dist proposed for theme_id "
                + str(theme_id)
                + " : "
                + str(dist_predicted_max)
            )
            predicted_result = dict_predicted[dist_predicted_max][theme_id]
            dict_evaluated_result[dist_predicted_max][theme_id] = predicted_result
            prop_dictionary[dist_predicted_max][theme_id]["formula"] = json.dumps(
                actual_aligner.get_formula(predicted_result["result"])
            )
            prop_dictionary[dist_predicted_max][theme_id][
                "evaluation"
            ] = Evaluation.TO_CHECK_4

    for theme_id, geom in dict_unchanged.items():
        result = {"result": geom}
        dict_evaluated_result[0][theme_id] = result
        prop_dictionary[0][theme_id]["evaluation"] = Evaluation.NO_CHANGE_6
        prop_dictionary[0][theme_id]["formula"] = json.dumps(
            actual_aligner.get_formula(result["result"])
        )
    return dict_evaluated_result, prop_dictionary


def check_equality(
    base_formula, actual_formula, threshold_area=5, threshold_percentage=1
):
    """
    function that checks if 2 formulas are equal (determined by business-logic)
    """
    # TODO: research naar aanduid_id 116448 (equality na 0.5m), 120194 (1m)
    # TODO: research and implementation of following ideas
    # TODO: refine equality comparison, make it more generic
    # ideas:
    # * If result_diff smaller than 0.x --> automatic update
    # * big polygons: If 'outer ring' has same formula (do net check inner side) -->
    #   automatic update
    # ** outer ring can be calculated: 1) negative buffer 2) original - buffered

    if base_formula is None or actual_formula is None:
        return False, Evaluation.NO_PREDICTION_5
    if (
        base_formula["reference_features"].keys()
        == actual_formula["reference_features"].keys()
    ):
        if base_formula["full"] and base_formula["full"]:
            return True, Evaluation.EQUALITY_FORMULA_GEOM_1

        equal_reference_features = True
        for key in base_formula["reference_features"].keys():
            if (
                (
                    base_formula["reference_features"][key]["full"]
                    == actual_formula["reference_features"][key]["full"]
                )
                or (
                    abs(
                        base_formula["reference_features"][key]["area"]
                        - actual_formula["reference_features"][key]["area"]
                    )
                    > threshold_area
                )
                or (
                    (
                        abs(
                            base_formula["reference_features"][key]["area"]
                            - actual_formula["reference_features"][key]["area"]
                        )
                        * 100
                        / base_formula["reference_features"][key]["area"]
                    )
                    > threshold_percentage
                )
            ):
                equal_reference_features = False
        if equal_reference_features:
            return True, Evaluation.EQUALITY_FORMULA_2
    if base_formula["full"] and base_formula["full"]:
        return True, Evaluation.EQUALITY_GEOM_3
    return False, Evaluation.NO_PREDICTION_5


class GRBActualLoader(GeoJsonLoader):
    def __init__(self, grb_type: GRBType, aligner, partition: int = 1000):
        super().__init__()
        self.aligner = aligner
        self.grb_type = grb_type
        self.part = partition
        self.data_dict_source["source"] = grb_type.value

    def load_data(self):
        if not self.aligner.dict_thematic:
            raise ValueError("Thematic data not loaded")
        geom_union = buffer_pos(self.aligner.get_thematic_union(), MAX_REFERENCE_BUFFER)
        collection, id_property = get_collection_grb_actual(
            grb_type=self.grb_type,
            geometry=geom_union,
            partition=self.part,
            crs=self.aligner.CRS,
        )
        self.id_property = id_property
        self.input = dict(collection)
        self.data_dict_source["version_date"] = datetime.now().strftime(date_format)
        self.aligner.logger.feedback_info(f"GRB downloaded: {self.grb_type}")
        return super().load_data()


class GRBFiscalParcelLoader(GeoJsonLoader):
    def __init__(self, year: str, aligner, partition=1000):
        super().__init__(_input=None, id_property=GRB_PARCEL_ID)
        self.aligner = aligner
        self.year = year
        self.part = partition
        self.data_dict_source["source"] = "Adpf"
        self.data_dict_source["version_date"] = datetime(int(year), 1, 1).strftime(
            date_format
        )

    def load_data(self):
        if not self.aligner.dict_thematic:
            raise ValueError("Thematic data not loaded")
        geom_union = buffer_pos(self.aligner.get_thematic_union(), MAX_REFERENCE_BUFFER)
        collection = get_collection_grb_fiscal_parcels(
            year=self.year,
            geometry=geom_union,
            partition=self.part,
            crs=self.aligner.CRS,
        )
        self.input = dict(collection)
        self.aligner.logger.feedback_info(f"Adpf downloaded for year: {self.year}")
        return super().load_data()



import numpy as np

from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import GRBActualLoader
from brdr.loader import GeoJsonFileLoader
from brdr.utils import dict_series_by_keys
from examples import show_map

# Press the green button in the gutter to run the script.
if __name__ == "__main__":
    """
    example to use the predictor-function to automatically predict which resulting
    geometries are interesting to look at (based on detection of breakpoints and
    relevant distances of 'no-change')
    """
    # Initiate an Aligner
    aligner = Aligner()
    # Load thematic data & reference data
    loader = GeoJsonFileLoader(
        "../tests/testdata/test_wanted_changes.geojson", "theme_id"
    )
    aligner.load_thematic_data(loader)
    loader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner)
    aligner.load_reference_data(loader)

    series = np.arange(0, 300, 10, dtype=int) / 100
    # predict which relevant distances are interesting to propose as resulting geometry
    dict_series, dict_predicted, diffs = aligner.predictor(
        relevant_distances=series, od_strategy=4, threshold_overlap_percentage=50
    )
    dict_predicted = dict_series_by_keys(dict_predicted)
    for key in dict_predicted:
        show_map(
            dict_predicted[key],
            {key: aligner.dict_thematic[key]},
            aligner.dict_reference,
        )



import os
import unittest

import numpy as np
from shapely import Point
from shapely import from_wkt
from shapely.geometry import Polygon
from shapely.geometry import shape

from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.enums import OpenbaarDomeinStrategy
from brdr.geometry_utils import buffer_neg_pos
from brdr.geometry_utils import grid_bounds
from brdr.loader import GeoJsonLoader
from brdr.typings import FeatureCollection
from brdr.typings import ProcessResult


class TestAligner(unittest.TestCase):
    def setUp(self):
        # Create a sample geometry for testing
        self.sample_aligner = Aligner()
        self.sample_geom = Polygon([(0, 0), (0, 10), (10, 10), (10, 0)])

    def test_buffer_neg_pos(self):
        # Check if the result of the _buffer_neg_pos gives an equal geometry
        out = buffer_neg_pos(self.sample_geom, 3)
        self.assertEqual(self.sample_geom, out)

    def test_grid_bounds_1(self):
        # Test _grid_bounds function
        delta = 1.0
        grid_partitions = grid_bounds(self.sample_geom, delta)

        # Check if the result is a list of Polygon objects
        self.assertIsInstance(grid_partitions, list)
        for partition in grid_partitions:
            self.assertIsInstance(partition, Polygon)

        # Add more specific tests based on your requirements

    def test_grid_bounds_2(self):
        # Test _grid_bounds function
        delta = 2.0
        grid_partitions = grid_bounds(self.sample_geom, delta)

        # Check if the result is a list of Polygon objects
        self.assertIsInstance(grid_partitions, list)
        for partition in grid_partitions:
            self.assertIsInstance(partition, Polygon)

    def test_export_results(self):
        aligner = Aligner()
        aligner.load_thematic_data_dict(
            {"theme_id_1": from_wkt("POLYGON ((0 0, 0 9, 5 10, 10 0, 0 0))")}
        )
        aligner.load_reference_data_dict(
            {"ref_id_1": from_wkt("POLYGON ((0 1, 0 10,8 10,10 1,0 1))")}
        )
        aligner.process_dict_thematic()
        path = "./tmp/"
        aligner.export_results(path=path)
        filenames = [f"{k}.geojson" for k in ProcessResult.__annotations__]
        for file_name in os.listdir(path):
            os.remove(path + file_name)
            assert file_name in filenames
        os.rmdir(path)

    def test_get_formula_full_intersection(self):
        # Test when intersection equals reference geometry
        key = "a"
        ref_dict = {key: self.sample_geom}
        self.sample_aligner.load_reference_data_dict(ref_dict)
        res = self.sample_aligner.get_formula(self.sample_geom, with_geom=True)
        self.assertTrue(res["full"])
        result = res["reference_features"][key]
        self.assertTrue(result["full"])
        self.assertEqual(result["percentage"], 100)

    def test_get_formula_partial_intersection(self):
        # Test when intersection is partial
        key = "a"
        ref_dict = {key: self.sample_geom.buffer(0.5)}
        self.sample_aligner.load_reference_data_dict(ref_dict)
        res = self.sample_aligner.get_formula(self.sample_geom, with_geom=True)
        self.assertFalse(res["full"])
        result = res["reference_features"][key]
        self.assertFalse(result["full"])
        self.assertGreater(result["percentage"], 0)
        self.assertLess(result["percentage"], 100)

    def test_process_geometry(self):
        # Test if processed geometry is equal to reference geometry
        key_ref = "a"
        ref_dict = {key_ref: self.sample_geom}
        self.sample_aligner.load_reference_data_dict(ref_dict)
        process_result = self.sample_aligner.process_geometry(
            self.sample_geom.buffer(0.5)
        )
        self.assertTrue(
            from_wkt(process_result["result"].wkt).equals(
                from_wkt(self.sample_geom.wkt)
            )
        )
        self.assertFalse(process_result["result_diff"].is_empty)

    def test_predictor(self):
        # Load thematic data & reference data
        thematic_dict = {"theme_id": from_wkt("POLYGON ((0 0, 0 9, 5 10, 10 0, 0 0))")}
        # ADD A REFERENCE POLYGON TO REFERENCE DICTIONARY
        reference_dict = {"ref_id": from_wkt("POLYGON ((0 1, 0 10,8 10,10 1,0 1))")}
        # LOAD THEMATIC DICTIONARY
        self.sample_aligner.load_thematic_data_dict(thematic_dict)
        # LOAD REFERENCE DICTIONARY
        self.sample_aligner.load_reference_data_dict(reference_dict)
        series = np.arange(0, 300, 10, dtype=int) / 100
        # predict which relevant distances are interesting to propose as resulting
        # geometry

        dict_series, dict_predicted, dict_diffs = self.sample_aligner.predictor(
            relevant_distances=series, od_strategy=4, threshold_overlap_percentage=50
        )
        self.assertEqual(len(dict_predicted[0]), len(thematic_dict))

    def test_load_reference_data_grb_actual_adp(self):
        thematic_dict = {
            "theme_id_1": from_wkt(
                "MultiPolygon (((174184.09476602054201066 171899.68933439542888664, "
                "174400.56834639035514556 171832.959863749332726, "
                "174388.65236948925303295 171770.99678386366576888, "
                "174182.10876987033407204 171836.13745758961886168, "
                "174184.88916448061354458 171873.07698598300339654, "
                "174184.09476602054201066 171899.68933439542888664)))"
            )
        }
        self.sample_aligner.load_thematic_data_dict(thematic_dict)
        # LOAD REFERENCE DICTIONARY
        self.sample_aligner.load_reference_data_grb_actual(
            grb_type=GRBType.ADP, partition=1000
        )
        self.assertGreater(len(self.sample_aligner.dict_reference), 0)

    def test_load_reference_data_grb_actual_gbg(self):
        thematic_dict = {
            "theme_id_1": from_wkt(
                "MultiPolygon (((174184.09476602054201066 171899.68933439542888664, "
                "174400.56834639035514556 171832.959863749332726, "
                "174388.65236948925303295 171770.99678386366576888, "
                "174182.10876987033407204 171836.13745758961886168, "
                "174184.88916448061354458 171873.07698598300339654, "
                "174184.09476602054201066 171899.68933439542888664)))"
            )
        }
        self.sample_aligner.load_thematic_data_dict(thematic_dict)
        # LOAD REFERENCE DICTIONARY
        self.sample_aligner.load_reference_data_grb_actual(
            grb_type=GRBType.GBG, partition=1000
        )
        self.assertGreater(len(self.sample_aligner.dict_reference), 0)

    def test_load_reference_data_grb_actual_knw(self):
        thematic_dict = {
            "theme_id_1": from_wkt(
                "MultiPolygon (((174184.09476602054201066 171899.68933439542888664, "
                "174400.56834639035514556 171832.959863749332726, "
                "174388.65236948925303295 171770.99678386366576888, "
                "174182.10876987033407204 171836.13745758961886168, "
                "174184.88916448061354458 171873.07698598300339654, "
                "174184.09476602054201066 171899.68933439542888664)))"
            )
        }
        self.sample_aligner.load_thematic_data_dict(thematic_dict)
        # LOAD REFERENCE DICTIONARY
        self.sample_aligner.load_reference_data_grb_actual(
            grb_type=GRBType.KNW, partition=1000
        )
        self.sample_aligner.process_dict_thematic()
        self.assertGreaterEqual(len(self.sample_aligner.dict_reference), 0)

    def test_all_od_strategies(self):
        # Load thematic data & reference data
        thematic_dict = {
            "theme_id_1": from_wkt("POLYGON ((0 0, 0 9, 5 10, 10 0, 0 0))")
        }
        # ADD A REFERENCE POLYGON TO REFERENCE DICTIONARY
        reference_dict = {"ref_id_1": from_wkt("POLYGON ((0 1, 0 10,8 10,10 1,0 1))")}
        # LOAD THEMATIC DICTIONARY
        self.sample_aligner.load_thematic_data_dict(thematic_dict)
        # LOAD REFERENCE DICTIONARY
        self.sample_aligner.load_reference_data_dict(reference_dict)
        for od_strategy in OpenbaarDomeinStrategy:
            process_result = self.sample_aligner.process_dict_thematic(
                relevant_distance=1,
                od_strategy=od_strategy,
                threshold_overlap_percentage=50,
            )
            self.assertEqual(len(process_result["theme_id_1"]), 6)

    def test_process_interior_ring(self):
        thematic_dict = {
            "theme_id_1": from_wkt(
                "MultiPolygon (((174370.67910432978533208 171012.2469546866195742, "
                "174461.24052877808571793 171002.71417316573206335, "
                "174429.46459037516615354 170869.25523187351063825, "
                "174373.85669817010057159 170894.6759825958579313, "
                "174369.09030740967136808 170896.6619787460367661, "
                "174370.67910432978533208 171012.2469546866195742),"
                "(174400.07184735251939856 170963.78864862219779752, "
                "174401.26344504262669943 170926.4519209987774957, "
                "174419.53460962430108339 170922.47992869839072227, "
                "174429.06739114515949041 170958.62505863170372322, "
                "174400.07184735251939856 170963.78864862219779752)))"
            )
        }
        self.sample_aligner.load_thematic_data_dict(thematic_dict)
        # LOAD REFERENCE DICTIONARY
        self.sample_aligner.load_reference_data_grb_actual(
            grb_type=GRBType.ADP, partition=1000
        )
        result_dict = self.sample_aligner.process_dict_thematic()
        self.assertEqual(len(result_dict), len(thematic_dict))

    def test_process_circle(self):
        geometry = Point(0, 0).buffer(3)
        thematic_dict = {"key": geometry}
        self.sample_aligner.load_thematic_data_dict(thematic_dict)
        # LOAD REFERENCE DICTIONARY
        self.sample_aligner.load_reference_data_grb_actual(
            grb_type=GRBType.ADP, partition=1000
        )
        results_dict = self.sample_aligner.process_dict_thematic()
        self.assertEqual(geometry, results_dict["key"]["result"])

    def test__prepare_thematic_data(self):
        aligner = Aligner()
        geojson: FeatureCollection = {
            "type": "FeatureCollection",
            "name": "theme",
            "crs": {
                "type": "name",
                "properties": {"name": "urn:ogc:def:crs:EPSG::31370"},
            },
            "features": [
                {
                    "type": "Feature",
                    "properties": {"fid": 4, "id": 4, "theme_identifier": "4"},
                    "geometry": {
                        "type": "MultiPolygon",
                        "coordinates": [
                            [
                                [
                                    [174647.924166895216331, 170964.980246312363306],
                                    [174693.204879119351972, 170943.531487890402786],
                                    [174696.382472959638108, 170930.82111252922914],
                                    [174678.905706838035258, 170901.428369506524177],
                                    [174660.634542256389977, 170861.708446502889274],
                                    [174641.56897921464406, 170820.399726579111302],
                                    [174593.905071610264713, 170839.46528962085722],
                                    [174614.559431572153699, 170881.568408004706725],
                                    [174628.064205393398879, 170926.054721768799936],
                                    [174647.924166895216331, 170964.980246312363306],
                                ]
                            ]
                        ],
                    },
                }
            ],
        }
        thematic_loader = GeoJsonLoader(_input=geojson, id_property="theme_identifier")
        aligner.dict_thematic, properties, source = thematic_loader.load_data()
        assert aligner.dict_thematic == {"4": shape(geojson["features"][0]["geometry"])}
        self.assertGreater(len(aligner.dict_thematic), 0)

    def test_get_results_as_dict(self):
        self.sample_aligner.load_thematic_data_dict(
            {
                "theme_id_1": from_wkt("POLYGON ((0 0, 0 9, 5 10, 10 0, 0 0))"),
                "theme_id_2": from_wkt("POLYGON ((0 0, 0 9, 5 10, 10 0, 0 0))"),
            }
        )
        self.sample_aligner.load_reference_data_dict(
            {"ref_id_1": from_wkt("POLYGON ((0 1, 0 10,8 10,10 1,0 1))")}
        )
        self.sample_aligner.process_dict_thematic()
        result = self.sample_aligner.get_results_as_dict()
        assert len(result) == 2

    def test_get_reference_as_geojson(self):
        self.sample_aligner.load_thematic_data_dict(
            {"theme_id_1": from_wkt("POLYGON ((0 0, 0 9, 5 10, 10 0, 0 0))")}
        )
        self.sample_aligner.load_reference_data_dict(
            {"ref_id_1": from_wkt("POLYGON ((0 1, 0 10,8 10,10 1,0 1))")}
        )
        self.sample_aligner.process_dict_thematic()
        self.sample_aligner.get_reference_as_geojson()

    def test_fully_aligned_input(self):
        aligned_shape = from_wkt("POLYGON ((0 0, 0 9, 5 10, 10 0, 0 0))")
        self.sample_aligner.load_thematic_data_dict({"theme_id_1": aligned_shape})
        self.sample_aligner.load_reference_data_dict({"ref_id_1": aligned_shape})
        result = self.sample_aligner.process_dict_thematic()
        assert result["theme_id_1"].get("result") == aligned_shape
        assert result["theme_id_1"].get("result_diff") == Polygon()
        assert result["theme_id_1"].get("result_diff_min") == Polygon()
        assert result["theme_id_1"].get("result_diff_plus") == Polygon()

    def test_fully_aligned_geojson_output(self):
        aligned_shape = from_wkt(
            "MultiPolygon (((173463.11530961000244133 174423.83310307000647299, "
            "173460.22633100001257844 174422.02316300000529736, "
            "173455.24681099998997524 174429.98009100000490434, "
            "173454.4299790000077337 174429.34482699999352917, "
            "173452.06690700000035577 174432.43058700000983663, "
            "173451.25743500000680797 174431.8672589999914635, "
            "173448.74844299998949282 174434.96249100001296028, "
            "173448.5809550000121817 174435.80485899999621324, "
            "173455.39772300000186078 174441.47852299999794923, "
            "173461.44169100001454353 174446.50898700000834651, "
            "173472.15932299999985844 174429.49919500001124106, "
            "173466.18524341000011191 174425.75641125999391079, "
            "173466.18524347001221031 174425.75641117000486702, "
            "173463.11530969000887126 174423.83310300001176074, "
            "173463.11530961000244133 174423.83310307000647299)))"
        )

        self.sample_aligner.load_thematic_data_dict({"theme_id_1": aligned_shape})
        self.sample_aligner.load_reference_data_dict({"ref_id_1": aligned_shape})
        self.sample_aligner.process_dict_thematic()
        fcs = self.sample_aligner.get_results_as_geojson(formula=True)
        assert fcs["result"]["features"][0]["properties"]["area"] > 0
        assert fcs["result_diff"]["features"][0]["properties"]["area"] == 0
        assert fcs["result_diff_min"]["features"][0]["properties"]["area"] == 0
        assert fcs["result_diff_plus"]["features"][0]["properties"]["area"] == 0



import logging
from datetime import date
from datetime import timedelta

import numpy as np

from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import GRBActualLoader
from brdr.grb import GRBFiscalParcelLoader
from brdr.grb import evaluate
from brdr.grb import get_geoms_affected_by_grb_change
from brdr.loader import DictLoader
from brdr.loader import GeoJsonLoader
from brdr.utils import get_oe_geojson_by_bbox

# This code shows an example how the aligner can be used inside a flow of
# parcel change detection:
# * it can be used to do a first alignment of the original features
#       (in this case, based on parcel version adpF2023 (1st January 2023)
# * it can be used to do a new alignment on the actual version of the parcels adp
# * it can be used to convert the geometries to a formula, to compare and
#       evaluate if equality is detected after alignement

counter_excluded = 0
# PARAMS
# =========
crs = "EPSG:31370"
limit = 10000
# bbox = "172800,170900,173000,171100"
bbox = "172000,172000,174000,174000"
# bbox = "170000,170000,175000,174900"
# bbox = "100000,195000,105000,195900"
# bbox = "150000,210000,155000,214900"
# bbox = "173500,173500,174000,174000" # example "aanduid_id" = 34195
base_year = "2022"
# relevant distance that is used to align the original geometries to the
# reference-polygons of the base-year
base_correction = 2
# geometries bigger than this, will be excluded
excluded_area = 10000
# series of relevant distance that is used to check if we can auto-align the geometries
# to the actual reference-polygons to get an 'equal' formula
# series = [
#     0,
#     0.5,
#     1,
#     1.5,
#     2,
# ]
series = np.arange(0, 200, 10, dtype=int) / 100
# BASE
# =====
# Initiate an Aligner to create a themeset that is base-referenced on a specific
# base_year
base_aligner = Aligner()
# Load the thematic data to evaluate
loader = GeoJsonLoader(_input=get_oe_geojson_by_bbox(bbox), id_property="aanduid_id")
# loader = DictLoader(get_oe_dict_by_ids(['554','1573','124023','1873','1782','1324']))
base_aligner.load_thematic_data(loader)

logging.info(
    "Number of OE-thematic features loaded into base-aligner: "
    + str(len(base_aligner.dict_thematic))
)
base_aligner.load_reference_data(
    GRBFiscalParcelLoader(year=base_year, aligner=base_aligner)
)

# Exclude objects bigger than specified area
keys_to_exclude = []
for key in base_aligner.dict_thematic:
    if base_aligner.dict_thematic[key].area > excluded_area:
        keys_to_exclude.append(key)
        counter_excluded = counter_excluded + 1
        logging.info(
            "geometrie excluded; bigger than " + str(excluded_area) + ": " + key
        )
for x in keys_to_exclude:
    del base_aligner.dict_thematic[x]


# Align the features to the base-GRB
base_process_result = base_aligner.process_dict_thematic(
    relevant_distance=base_correction
)

thematic_dict_formula = {}
thematic_dict_result = {}
i = 0
for key in base_process_result:
    i = i + 1
    thematic_dict_result[key] = base_process_result[key]["result"]
    thematic_dict_formula[key] = base_aligner.get_formula(thematic_dict_result[key])
    if i > 500:
        break

base_aligner_result = Aligner()
base_aligner_result.load_thematic_data(DictLoader(thematic_dict_result))
dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
    aligner=base_aligner_result,
    grb_type=GRBType.ADP,
    date_start=date.today() - timedelta(days=365),
    date_end=date.today(),
    one_by_one=False,
)

logging.info(
    "Number of possible affected OE-thematic during timespan: "
    + str(len(dict_affected))
)

# ACTUAL
# Initiate a Aligner to reference thematic features to the actual borders
# ================================================================================

# Initiate an Aligner to reference thematic features to the actual borders
actual_aligner = Aligner()
loader = DictLoader(dict_affected)
actual_aligner.load_thematic_data(loader)
loader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=actual_aligner)
actual_aligner.load_reference_data(loader)

# LOOP AND PROCESS ALL POSSIBLE AFFECTED FEATURES
# =================================================
series = np.arange(0, 200, 10, dtype=int) / 100
dict_series, dict_predicted, diffs_dict = actual_aligner.predictor(series)
dict_evaluated_result, prop_dictionary = evaluate(
    actual_aligner,
    dict_series,
    dict_predicted,
    thematic_dict_formula,
    threshold_area=5,
    threshold_percentage=1,
    dict_unchanged=dict_unchanged,
)
counter_equality = 0
counter_equality_by_alignment = 0
counter_difference = 0
for theme_id in dict_affected:
    for dist in series:
        if "evaluation" in prop_dictionary[dist][theme_id].keys():
            ev = prop_dictionary[dist][theme_id]["evaluation"]
            if ev.startswith("equal") and dist == 0:
                counter_equality = counter_equality + 1
            elif ev.startswith("equal") and dist > 0:
                counter_equality_by_alignment = counter_equality_by_alignment + 1
            else:
                counter_difference = counter_difference + 1
            break

logging.info(
    "Features: "
    + str(len(dict_affected))
    + "//Equality: "
    + str(counter_equality)
    + "//Equality by alignment: "
    + str(counter_equality_by_alignment)
    + "//Difference: "
    + str(counter_difference)
    + "//Excluded: "
    + str(counter_excluded)
)



from datetime import date

import numpy as np

from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import (
    get_geoms_affected_by_grb_change,
    evaluate,
    GRBFiscalParcelLoader,
    GRBActualLoader,
)
from brdr.loader import DictLoader
from brdr.utils import get_series_geojson_dict, get_oe_dict_by_ids

# dict_theme = get_oe_dict_by_ids([125610,148305,127615,122316,120153,124699,115489,
# 120288,120387,124762,148143,116141])
dict_theme = get_oe_dict_by_ids([10047, 10048, 10049, 10050, 10051, 10056])
print(dict_theme)

base_aligner = Aligner()
base_aligner.load_thematic_data(DictLoader(dict_theme))
base_year = "2022"
base_aligner.load_reference_data(
    GRBFiscalParcelLoader(year=base_year, aligner=base_aligner)
)
base_process_result = base_aligner.process_dict_thematic(relevant_distance=2)
thematic_dict_formula = {}
thematic_dict_result = {}
for key in base_process_result:
    thematic_dict_result[key] = base_process_result[key]["result"]
    thematic_dict_formula[key] = base_aligner.get_formula(thematic_dict_result[key])
base_aligner_result = Aligner()
base_aligner_result.load_thematic_data(DictLoader(thematic_dict_result))
dict_affected, dict_unchanged = get_geoms_affected_by_grb_change(
    base_aligner_result,
    grb_type=GRBType.ADP,
    date_start=date(2022, 1, 1),
    date_end=date.today(),
    one_by_one=False,
)
if dict_affected == {}:
    print("No affected dicts")
    exit()

actual_aligner = Aligner()
loader = DictLoader(dict_affected)
actual_aligner.load_thematic_data(loader)
loader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=actual_aligner)
actual_aligner.load_reference_data(loader)
series = np.arange(0, 200, 10, dtype=int) / 100
dict_series, dict_predicted, diffs_dict = actual_aligner.predictor(series)

# diffs_dict=merge_diffs_dict(diffs_dict)

dict_evaluated, prop_dictionary = evaluate(
    actual_aligner,
    dict_series,
    dict_predicted,
    thematic_dict_formula,
    threshold_area=5,
    threshold_percentage=1,
    dict_unchanged=dict_unchanged,
)

fc = get_series_geojson_dict(
    dict_evaluated,
    crs=actual_aligner.CRS,
    id_field=actual_aligner.name_thematic_id,
    series_prop_dict=prop_dictionary,
)
for feature in fc["result"]["features"]:
    print(feature["properties"]["evaluation"])



from brdr.aligner import Aligner
from brdr.enums import OpenbaarDomeinStrategy, GRBType
from brdr.utils import diffs_from_dict_series
from examples import plot_series
from examples import show_map

if __name__ == "__main__":
    # Initiate brdr
    aligner = Aligner()
    # Load thematic data
    aligner.load_thematic_data_file(
        "../tests/testdata/themelayer_referenced.geojson", "id_theme"
    )

    # Use GRB adp-parcels as reference polygons adp= administratieve percelen
    aligner.load_reference_data_grb_actual(grb_type=GRBType.ADP, partition=1000)
    # alternative reference poly
    # # Use GRB-gbg (buildings), gbg= gebouw aan de grond
    # x.load_reference_data_grb_actual('gbg')

    # Use local data
    # x.load_reference_data_file(
    # "../tests/testdata/reference_leuven.geojson", 'capakey'
    # )

    # Example how to use the Aligner
    rel_dist = 10
    dict_results_by_distance = {
        rel_dist: aligner.process_dict_thematic(
            relevant_distance=rel_dist,
            od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_SINGLE_SIDE,
        )
    }
    aligner.export_results("output/")
    show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)

    rel_dist = 6
    dict_results_by_distance = {
        rel_dist: aligner.process_dict_thematic(
            relevant_distance=rel_dist, od_strategy=OpenbaarDomeinStrategy.SNAP_ALL_SIDE
        )
    }
    aligner.export_results("output/")
    show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)
    # for key in r:
    #     x.get_formula(r[key])

    # Example how to use a series (for histogram)
    series = [0.1, 0.2, 0.3, 0.4, 0.5, 1, 2]
    dict_series = aligner.process_series(series, 2, 50)
    resulting_areas = diffs_from_dict_series(dict_series, aligner.dict_thematic)
    plot_series(series, resulting_areas)

    # Example how to use the Aligner with threshold_overlap_percentage=-1 (original
    # border will be used for cases where relevant zones cannot be used for
    # determination)
    rel_dist = 6
    dict_results_by_distance = {
        rel_dist: aligner.process_dict_thematic(
            relevant_distance=rel_dist,
            od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE,
            threshold_overlap_percentage=-1,
        )
    }
    aligner.export_results("output/")
    show_map(dict_results_by_distance, aligner.dict_thematic, aligner.dict_reference)
